import argparse
import matplotlib.pyplot as plt
import numpy as np
import sys, os
from pyproj import Proj
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from netCDF4 import Dataset

sys.path.insert(0, "static/")
from ABIcmaps import irw_ctc, wv_cimss_r
from lightningcast import utils
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER


def abi_img(filename, extents=[-999], outdir=os.environ["PWD"] + "/"):

    nc = Dataset(filename, "r")

    radiance = nc.variables["Rad"][:]
    basefile = os.path.basename(filename)
    baseparts = basefile.split("_")
    channel = baseparts[1][-2:]
    goes_position = nc.orbital_slot
    scene_id = nc.scene_id
    if int(channel) <= 6:
        # convert vis radiance to reflectance
        result = radiance * nc.variables["kappa0"][:]
        xlab_suffix = "reflectance []"
        vmin = 0
        vmax = 1
        tickrange = np.arange(0, 1.01, 0.1)
        cmap = plt.get_cmap("Greys_r")
    else:
        # convert IR radiance to brightness temperature
        result = (
            nc.variables["planck_fk2"][:]
            / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
            - nc.variables["planck_bc1"][:]
        ) / nc.variables["planck_bc2"][:]
        xlab_suffix = "brightness temperature [K]"
        if channel in "08,09,10":
            vmin = 163
            vmax = 328
            tickrange = np.arange(163, 324, 20, dtype=int)
            cmap = wv_cimss_r()
        elif channel in "11,12,13,14,15,16":
            vmin = 180
            vmax = 310
            tickrange = np.arange(180, 311, 10, dtype=int)
            cmap = irw_ctc()
        # elif(channel == '07'):

    satheight = nc.variables["goes_imager_projection"].getncattr(
        "perspective_point_height"
    )
    lon0 = nc.variables["goes_imager_projection"].getncattr(
        "longitude_of_projection_origin"
    )
    sweep = nc.variables["goes_imager_projection"].getncattr("sweep_angle_axis")
    timestamp = getattr(nc, "time_coverage_start")[0:16]
    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    x = nc.variables["x"][:] * satheight
    y = nc.variables["y"][:] * satheight
    nc.close()

    p = Proj(proj="geos", h=satheight, lon_0=lon0, sweep=sweep)
    xx, yy = np.meshgrid(x, y)
    projection = ccrs.Geostationary(
        central_longitude=lon0, sweep_axis=sweep, satellite_height=satheight
    )
    lons, lats = p(xx, yy, inverse=True)
    lons = np.ma.filled(np.ma.fix_invalid(lons))
    lats = np.ma.filled(np.ma.fix_invalid(lats))
    result = np.ma.filled(np.ma.fix_invalid(result))
    del (x, y, xx, yy)
    print(type(lons), type(lats), type(result))

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(1, 1, 1, projection=projection)
    if extents[0] == -999:
        ax.set_extent([lons[-1, 0], lons[0, -1], lats[-1, 0], lats[0, -1]])
    else:
        ax.set_extent(extents)
    ax.add_feature(cfeature.BORDERS)
    ax.add_feature(cfeature.COASTLINE)
    state_borders = cfeature.NaturalEarthFeature(
        category="cultural",
        name="admin_1_states_provinces_lakes",
        scale="50m",
        facecolor="none",
    )
    ax.add_feature(state_borders, edgecolor="black")

    pcimg = ax.pcolormesh(
        lons,
        lats,
        result,
        vmin=vmin,
        vmax=vmax,
        transform=ccrs.PlateCarree(),
        cmap=cmap,
    )
    cbar = fig.colorbar(
        pcimg,
        orientation="horizontal",
        pad=0.04,
        shrink=0.80,
        extendfrac=0,
        drawedges=0,
        ticks=tickrange,
    )
    cbar.set_label("Channel " + channel + " " + xlab_suffix)
    cbar.ax.tick_params(labelsize=10)
    ax.set_title((" ").join([goes_position, scene_id, timestamp, "UTC"]), fontsize=12)

    # grid lines (this doesn't work with geos projection! >:-( )
    # gl = ax.gridlines(crs=ccrs.PlateCarree(), linewidth=1, color='cyan', alpha=0.5, linestyle='--', draw_labels=True)
    # gl.xlabels_top = False
    # gl.ylabels_left = False
    # gl.ylabels_right=True
    # gl.xlines = True
    # gl.xlocator = mticker.FixedLocator([120, 140, 160, 180, -160, -140, -120])
    # gl.ylocator = mticker.FixedLocator([0, 20, 40, 60])
    # gl.xformatter = LONGITUDE_FORMATTER
    # gl.yformatter = LATITUDE_FORMATTER
    # gl.xlabel_style = {'color': 'red', 'weight': 'bold'}

    # save
    utils.mkdir_p(outdir)
    fig.savefig(
        os.path.join(
            outdir,
            ("_").join(["C" + channel, scene_id, timestamp.replace(":", "") + ".png"]),
        ),
        bbox_inches="tight",
    )
    plt.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("file", help="ABI L1b file", type=str, nargs=1)
    parser.add_argument(
        "-e",
        "--extents",
        help="lat/lon extents [SWlon, NElon, SWlat, NElat]. Default is full image.",
        default=[-999],
        type=np.float32,
        nargs=4,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="Root outdir; Default = $PWD",
        default=[os.environ["PWD"] + "/"],
        type=str,
        nargs=1,
    )

    args = parser.parse_args()

    abi_img(args.file[0], args.extents, args.outdir[0])
