import numpy as np
import os, sys
import zarr
from datetime import datetime, timedelta
from netCDF4 import Dataset
import glob
import collect_data
import time

t0 = time.time()

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

# halfsize
hs = 32
hrrr_hs = 32

chunklen = 20000  # keep this small enough to keep small RAM footprint
chunksize = 300  # how to chunk data in outzarr
chunks = (chunksize, hs * 2, hs * 2, 1)

ghrsst_dir = "/iliad/grain/common/oconus_ps/GHRSST_5km/"
hrrr_dir = "/data/jcintineo/HRRR/remapped/"

inzarr = "/apollo/grain/jcintineo/AI/OCONUS/pltg_8dates.zarr"
outzarr = "/apollo/grain/jcintineo/AI/OCONUS/pltg_8dates.zarr"
ds = zarr.open(inzarr, "r")

dates = ds.ABI_date[:]
times = ds.ABI_time[:]
lats = ds.lat[:]
lons = ds.lon[:]
abix = ds.ABIx[:]
abiy = ds.ABIy[:]

varnames = [
    "u925_arr",
    "v925_arr",
    "u850_arr",
    "v850_arr",
    "u700_arr",
    "v700_arr",
    "u500_arr",
    "v500_arr",
    "mlcape_arr",
    "SST_arr",
]
hrrrnames = varnames[0:-1]
variables = {}
for var in varnames:
    variables[var] = np.zeros((chunklen, hs * 2, hs * 2, 1))

hrrr_files = {}
sst_files = {}

idx = 0
print("new chunk")
prev_date = dates[0]
for i in range(len(dates)):  # for each sample

    if dates[i] != prev_date:  # reset on each day to save memory
        hrrr_files = {}
        sst_files = {}
    prev_date = dates[i]

    hrrr_dt = datetime.strptime(
        str(dates[i]) + str(times[i]).zfill(4), "%Y%m%d%H%M"
    ) - timedelta(
        hours=1
    )  # to simulate real-time
    dt = datetime.strptime(str(dates[i]), "%Y%m%d") - timedelta(
        days=1
    )  # to simulate real-time

    # GHRSST
    dkey = dt.strftime("%Y%m%d")
    # find and open sst file
    if dkey not in sst_files:
        try:
            ncfile = glob.glob(os.path.join(ghrsst_dir, dkey + "*.nc"))[0]
        except IndexError:
            dt = datetime.strptime(str(dates[i]), "%Y%m%d") - timedelta(
                days=2
            )  # try 2 days ago
            ncfile = glob.glob(
                os.path.join(ghrsst_dir, dt.strftime("%Y%m%d") + "*.nc")
            )[0]
        print("opening " + ncfile)
        sst_nc = Dataset(ncfile, "r")
        try:
            sst_lat
        except NameError:
            sst_lat = sst_nc.variables["lat"][:]
            sst_lon = sst_nc.variables["lon"][:]
        sst_data = sst_nc.variables["analysed_sst"][0, :, :]
        sst_files[dkey] = sst_data

    sst_x = (np.abs(sst_lon - lons[i])).argmin()
    sst_y = (np.abs(sst_lat - lats[i])).argmin()

    variables["SST_arr"][idx, ..., 0] = sst_files[dkey][
        sst_y - hs : sst_y + hs, sst_x - hs : sst_x + hs
    ]

    # HRRR fields
    hrrr_hash = hrrr_dt.strftime("%Y%m%d_%H")
    hrrr_data = {
        "mlcape": -1,
        "u925": -1,
        "v925": -1,
        "u850": -1,
        "v850": -1,
        "u700": -1,
        "v700": -1,
        "u500": -1,
        "v500": -1,
    }
    y = abiy[i]
    x = abix[i]
    if hrrr_hash not in hrrr_files:
        ncfile = glob.glob(
            os.path.join(
                hrrr_dir, hrrr_hash[0:8], "HRRR_remapped_" + hrrr_hash + "0000.nc*"
            )
        )[0]
        print("opening " + ncfile)
        status = collect_data.read_netcdf(ncfile, data=hrrr_data)
        hrrr_files[hrrr_hash] = hrrr_data
    # now add data to variables
    for var in hrrrnames:
        vn = var.split("_")[0]
        try:
            variables[var][idx, ..., 0] = hrrr_files[hrrr_hash][vn][
                y - hs : y + hs, x - hs : x + hs
            ]
        except ValueError:
            variables[var][idx, ..., 0] = np.zeros((hs * 2, hs * 2)) - 999

    # save into zarr file
    idx += 1
    if idx == chunklen or i == (len(dates) - 1):
        print("new chunk")
        for var in varnames:
            zvar = os.path.join(outzarr, var)
            if os.path.isdir(zvar):
                # existing data
                z1 = zarr.open(zvar, mode="a")
                z1.append(variables[var][0:idx], axis=0)
            else:
                # new data
                # print(variables[var].shape)
                z1 = zarr.open(
                    zvar,
                    shape=variables[var][0:idx].shape,
                    chunks=chunks,
                    dtype=variables[var].dtype,
                    mode="a",
                )
                z1[:] = variables[var][0:idx]

        # now reset for new chunklen
        idx = 0
        variables = {}
        for var in varnames:
            variables[var] = np.zeros((chunklen, hs * 2, hs * 2, 1))


print((time.time() - t0) / 3600, "hours")
