import numpy as np
import os, sys
import glob
import netCDF4
from datetime import datetime, timedelta
import time
import collections
import multiprocessing as mp
from lightningcast.predict_ltg import get_abi_data
import itertools
from PIL import Image
from lightningcast import utils

t0 = time.time()

# Directories and variable names. Edit if necessary.
goes = "goes16"
position = "goes_east" if (goes == "goes16") else "goes_west"

outroot = f"/ships19/grain/probsevere/LC/bitmaps/{goes}"

abiroot = f"/arcdata/goes/grb/{goes}/"
abi_channels = ["CH02", "CH04", "CH05", "CH06", "CH09", "CH10", "CH11", "CH13", "CH15"]

targetroot = f"/apollo/grain/saved_probsevere_data/lightning/"
target_var = "FED_accum_60min_2km"

glmroot = targetroot
glmvar = "flash_extent_density"

climo_file = "/ships19/grain/probsevere/LC/ltg_climo_daily.nc"
climo_var = "HRAC_COM_FR"
climo_fac = 4  # 2 = 4-km res, 4 = 8-km res
do_climo = False

sfctype_file = "/ships19/grain/probsevere/LC/pixel_surface_type.nc"
sfctype_var = "pixel_surface_type"  # 2-km resolution
landmask_var = "LANDMASK"
do_landsea_mask = False

bsinfo = {}  # bytescaling info
bsinfo["CH01"] = {"vmin": 0, "vmax": 1}
bsinfo["CH02"] = {"vmin": 0, "vmax": 1}
bsinfo["CH03"] = {"vmin": 0, "vmax": 1}
bsinfo["CH04"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH05"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH06"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH07"] = {"vmin": 220, "vmax": 360}
bsinfo["CH08"] = {"vmin": 200, "vmax": 270}
bsinfo["CH09"] = {"vmin": 200, "vmax": 275}
bsinfo["CH10"] = {"vmin": 200, "vmax": 280}
bsinfo["CH11"] = {"vmin": 190, "vmax": 320}
bsinfo["CH12"] = {"vmin": 190, "vmax": 320}
bsinfo["CH13"] = {"vmin": 190, "vmax": 320}
bsinfo["CH14"] = {"vmin": 190, "vmax": 320}
bsinfo["CH15"] = {"vmin": 190, "vmax": 320}
bsinfo["CH16"] = {"vmin": 190, "vmax": 300}
bsinfo["FED"] = {"vmin": 0, "vmax": 255}
bsinfo[target_var] = {"vmin": 0, "vmax": 255}
bsinfo["HRAC_COM_FR"] = {"vmin": 0, "vmax": 0.24}  # mean = 0.023433; std = 0.030576
bsinfo["LANDMASK"] = {"vmin": 0, "vmax": 1}

# Edit patch info if necessary
patchlen = 128  # pixels, using 2-km data as reference
hs = patchlen // 2
patch_npix = patchlen**2

ny, nx = (1500, 2500)  # hard-coded for targets

# Edit thresholds for reflectivity (i.e., non-severe patches)
target_thresh = 1
# target_frac_thresh = 0
# ------------------------------------------------------------------------------------------------
def write_bmp(data_arr, vmin, vmax, outdir, prefix, metadata_dict={}):
    """
    Take in a numpy array and write out a .bmp
    to {outdir}. Optionally, write a metadata to {outdir}.
    h/t Tony Wimmers
    """
    utils.mkdir_p(outdir)

    DataImage = np.round((data_arr - vmin) / (vmax - vmin) * 255.9999)
    DataImage[DataImage < 0] = 0
    DataImage[DataImage > 255] = 255
    # DataImage[np.isnan(DataImage)] = 0 How we should handle bad or missing data?

    bmpArr = Image.fromarray(DataImage.astype(np.uint8))
    bmpArr.save(f"{outdir}/{prefix}.bmp")

    # metadata file
    if len(metadata_dict):
        meta_file = open(f"{outdir}/{prefix}.txt", "w")
        for key, val in metadata_dict.items():
            meta_file.write(f"{key}: {val}\n")
        meta_file.close()


# ------------------------------------------------------------------------------------------------
def get_glm_data(dt, glmroot, glmvar=glmvar):

    try:
        glmfile = glob.glob(
            dt.strftime(f"{glmroot}/%Y/%Y%m%d/GLM/{position}/agg/%Y%m%d-%H%M*.netcdf")
        )[0]
    except IndexError:
        for mmm in [-1, 1, -2, 2, -3, 3]:
            tmpdt = dt + timedelta(minutes=mmm)
            try:
                glmfile = glob.glob(
                    tmpdt.strftime(
                        f"{glmroot}/%Y/%Y%m%d/GLM/{position}/agg/%Y%m%d-%H%M*.netcdf"
                    )
                )[0]
            except IndexError:
                if mmm == 3:
                    raise

    glmdata = utils.read_data(glmfile, glmvar)
    return glmdata


# ------------------------------------------------------------------------------------------------
def create_additional_bmps_from_existing_samples(list_of_files):

    if do_climo:
        # get the lightning climo data
        climo_data = utils.read_data(climo_file, climo_var)

    if do_landsea_mask:
        # get/make the landsea mask
        pixel_surface_type = utils.read_data(sfctype_file, sfctype_var)
        landsea_mask = np.copy(pixel_surface_type)
        del pixel_surface_type
        landsea_mask[landsea_mask > 0] = 1

    for sample_file in list_of_files:
        # Note that outpath for new bmps is based off outdir of input files!!
        outpath = os.path.dirname(
            os.path.dirname(sample_file)
        )  # 2 dirnames removes variable dir.
        basefile = os.path.basename(sample_file)
        prefix = basefile[0:-4]  # remove .bmp
        # 20190603-2301_Y1216-X1728.bmp  2-km Y and X "center point" of patch (on CONUS or PACUS fixed grid)
        y = int(basefile.split("_Y")[1].split("-X")[0])
        x = int(basefile.split("-X")[1].split(".")[0])

        if do_climo:
            # make lightning climo bmp
            climo_patch = climo_data[
                doy,
                (y - hs) // climo_fac : (y + hs) // climo_fac,
                (x - hs) // climo_fac : (x + hs) // climo_fac,
            ]
            vmin = bsinfo[climo_var]["vmin"]
            vmax = bsinfo[climo_var]["vmax"]
            outdir = f"{outpath}/{climo_var}/"
            write_bmp(climo_patch, vmin=vmin, vmax=vmax, prefix=prefix, outdir=outdir)

        if do_landsea_mask:
            # make landsea mask bmp
            mask_patch = landsea_mask[y - hs : y + hs, x - hs : x + hs]
            vmin = bsinfo[landmask_var]["vmin"]
            vmax = bsinfo[landmask_var]["vmax"]
            outdir = f"{outpath}/{landmask_var}/"
            write_bmp(mask_patch, vmin=vmin, vmax=vmax, prefix=prefix, outdir=outdir)


# ------------------------------------------------------------------------------------------------
def create_predictor_and_target_bmps(list_of_files):
    cand_Y = np.arange(hs, ny - hs).astype(np.int32)
    cand_X = np.arange(hs, nx - hs).astype(np.int32)

    if do_landsea_mask:
        # get/make the landsea mask
        pixel_surface_type = utils.read_data(sfctype_file, sfctype_var)
        landsea_mask = np.copy(pixel_surface_type)
        del pixel_surface_type
        landsea_mask[landsea_mask > 0] = 1

    if do_climo:
        # get the lightning climo data
        climo_data = utils.read_data(climo_file, climo_var)

    for target_file in list_of_files:  # foreach datetime
        dt = datetime.strptime(os.path.basename(target_file)[0:13], "%Y%m%d-%H%M")

        abidt = dt - timedelta(
            minutes=60
        )  # N.B. The timestamp for the GLM/target data is the **END TIME** of the accumulation
        doy = int(abidt.strftime("%j")) - 1
        if doy > 364:
            doy = 364  # account for leap year and the fact that dataset uses 0-based index

        # get a dict of selected abi channels
        abi_data = {}
        cont = False
        for ch in abi_channels:
            try:
                abi_data[ch] = get_abi_data(abidt, ch[2:], abiroot, "RadC")
            except IndexError:
                cont = True
                break
        if cont:
            continue

        # get the GLM FED data
        try:
            fed_data = get_glm_data(abidt, glmroot)
        except IndexError:
            print("Couldn't find GLM data near ", abidt)
            continue  # move on to next target_file if we can't find FED near abidt

        # get the target data
        target_data = utils.read_data(target_file, target_var)

        # March through the target image and make bmps where we've met the criteria
        positive_centers = []
        for y in range(hs, ny - hs, patchlen):  # at 2-km
            for x in range(hs, nx - hs, patchlen):  # at 2-km

                target_patch = target_data[y - hs : y + hs, x - hs : x + hs]

                # If we've met the target criteria, make bmps
                if np.max(target_patch) >= 2:
                    sub = "2fl_odd"
                    positive_centers.append((y, x))
                    prefix = abidt.strftime(
                        f"%Y%m%d-%H%M_Y{y}-X{x}"
                    )  # use the prediction start time as the timestamp

                    # make target bmp
                    outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/{sub}/{target_var}/")
                    write_bmp(
                        target_patch,
                        vmin=bsinfo[target_var]["vmin"],
                        vmax=bsinfo[target_var]["vmax"],
                        prefix=prefix,
                        outdir=outdir,
                    )

                    # make ABI bmps
                    for ch in abi_channels:
                        if ch == "CH02":
                            abi_patch = abi_data[ch][
                                (y - hs) * 4 : (y + hs) * 4, (x - hs) * 4 : (x + hs) * 4
                            ]
                        elif ch in ["CH01", "CH03", "CH05"]:
                            abi_patch = abi_data[ch][
                                (y - hs) * 2 : (y + hs) * 2, (x - hs) * 2 : (x + hs) * 2
                            ]
                        else:
                            abi_patch = abi_data[ch][y - hs : y + hs, x - hs : x + hs]

                        vmin = bsinfo[ch]["vmin"]
                        vmax = bsinfo[ch]["vmax"]
                        outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/{sub}/{ch}/")
                        write_bmp(
                            abi_patch,
                            vmin=vmin,
                            vmax=vmax,
                            prefix=prefix,
                            outdir=outdir,
                        )

                    # make GLM FED bmp
                    glm_patch = fed_data[y - hs : y + hs, x - hs : x + hs]
                    outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/{sub}/FED/")
                    write_bmp(
                        glm_patch,
                        vmin=bsinfo["FED"]["vmin"],
                        vmax=bsinfo["FED"]["vmax"],
                        prefix=prefix,
                        outdir=outdir,
                    )

                    if do_climo:
                        # make lightning climo bmp
                        climo_patch = climo_data[
                            doy,
                            (y - hs) // climo_fac : (y + hs) // climo_fac,
                            (x - hs) // climo_fac : (x + hs) // climo_fac,
                        ]
                        vmin = bsinfo[climo_var]["vmin"]
                        vmax = bsinfo[climo_var]["vmax"]
                        outdir = abidt.strftime(
                            f"{outroot}/%Y/%Y%m%d/{sub}/{climo_var}/"
                        )
                        write_bmp(
                            climo_patch,
                            vmin=vmin,
                            vmax=vmax,
                            prefix=prefix,
                            outdir=outdir,
                        )

                    if do_landsea_mask:
                        # make landsea mask bmp
                        mask_patch = landsea_mask[y - hs : y + hs, x - hs : x + hs]
                        vmin = bsinfo[landmask_var]["vmin"]
                        vmax = bsinfo[landmask_var]["vmax"]
                        outdir = abidt.strftime(
                            f"{outroot}/%Y/%Y%m%d/{sub}/{landmask_var}/"
                        )
                        write_bmp(
                            mask_patch,
                            vmin=vmin,
                            vmax=vmax,
                            prefix=prefix,
                            outdir=outdir,
                        )

        ##Now try to find the same amount of negative patches
        # cand_Y = np.arange(hs,ny-hs).astype(np.int32)
        # cand_X = np.arange(hs,nx-hs).astype(np.int32)
        # for _ in positive_centers:
        #  #select 10 random candidates for centers. If none of them meet our criteria, then just move on
        #  Y = np.random.choice(cand_Y,10); X = np.random.choice(cand_X,10)
        #  for jj in range(len(Y)):  #Only 1 full iteration, since we break when condition is met
        #    y = Y[jj]; x = X[jj]
        #    target_patch = target_data[y-hs:y+hs,x-hs:x+hs]
        #    if(np.max(target_patch) < 1):
        #      prefix = abidt.strftime(f"%Y%m%d-%H%M_Y{y}-X{x}")

        #      #make target bmp
        #      outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/0fl/{target_var}/")
        #      write_bmp(target_patch, vmin=bsinfo[target_var]['vmin'], vmax=bsinfo[target_var]['vmax'], prefix=prefix, outdir=outdir)

        #      #make ABI bmps
        #      for ch in abi_channels:
        #        if(ch == 'CH02'):
        #          abi_patch = abi_data[ch][(y-hs)*4:(y+hs)*4, (x-hs)*4:(x+hs)*4]
        #        elif(ch in ['CH01','CH03','CH05']):
        #          abi_patch = abi_data[ch][(y-hs)*2:(y+hs)*2, (x-hs)*2:(x+hs)*2]
        #        else:
        #          abi_patch = abi_data[ch][y-hs:y+hs, x-hs:x+hs]

        #        vmin = bsinfo[ch]['vmin']; vmax = bsinfo[ch]['vmax']
        #        outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/0fl/{ch}/")
        #        write_bmp(abi_patch, vmin=vmin, vmax=vmax, prefix=prefix, outdir=outdir)

        #      #make GLM FED bmp
        #      glm_patch = fed_data[y-hs:y+hs, x-hs:x+hs]
        #      outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/0fl/FED/")
        #      write_bmp(glm_patch, vmin=bsinfo['FED']['vmin'], vmax=bsinfo['FED']['vmax'], prefix=prefix, outdir=outdir)

        #      if(do_climo):
        #        #make lightning climo bmp
        #        climo_patch = climo_data[doy, (y-hs)//climo_fac:(y+hs)//climo_fac, (x-hs)//climo_fac:(x+hs)//climo_fac]
        #        vmin = bsinfo[climo_var]['vmin']; vmax = bsinfo[climo_var]['vmax']
        #        outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/0fl/{climo_var}/")
        #        write_bmp(climo_patch, vmin=vmin, vmax=vmax, prefix=prefix, outdir=outdir)

        #      if(do_landsea_mask):
        #        #make landsea mask bmp
        #        mask_patch = landsea_mask[y-hs:y+hs, x-hs:x+hs]
        #        vmin = bsinfo[landmask_var]['vmin']; vmax = bsinfo[landmask_var]['vmax']
        #        outdir = abidt.strftime(f"{outroot}/%Y/%Y%m%d/0fl/{landmask_var}/")
        #        write_bmp(mask_patch, vmin=vmin, vmax=vmax, prefix=prefix, outdir=outdir)

        #      break #move on to next positive center

        print(f'processed {abidt.strftime("%Y%m%d-%H%M%S")}')


#####################################################

if __name__ == "__main__":

    # target netcdfs
    files = np.sort(
        glob.glob(
            f"{targetroot}/2019/2019???[1,3,5,7,9]/GLM/{position}/{target_var}/*.netcdf"
        )
    )
    files = files[::3]  # every X files

    # already-existing samples
    # files = np.sort(glob.glob(f"/ships19/grain/probsevere/LC/bitmaps/goes16/2019/2019*/?fl/CH13/*.bmp"))

    nthreads = 40
    pool = mp.Pool(processes=nthreads)

    interval = int(np.ceil(len(files) / nthreads))
    list_of_lists = []
    for n in range(nthreads):
        list_of_lists.append(files[n * interval : n * interval + interval])

    print(len(files))

    # target_files=list(np.arange(100))
    print(len(list_of_lists), interval)
    # sys.exit()

    # for new samples
    pool.map(create_predictor_and_target_bmps, list_of_lists)

    # for appending to existing samples
    # pool.map(create_additional_bmps_from_existing_samples,list_of_lists)

    # results = pool.starmap(create_predictor_and_target_bmps,zip(list_of_lists,itertools.repeat(1)))
    pool.close()
    # pool.join()

    print("Time elapsed:", (time.time() - t0) / 60, " minutes")
