import numpy as np
import json
import os, sys
import netCDF4
import argparse
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import unicodedata
from shapely.geometry import Point, Polygon, MultiPolygon, shape
import datetime
import pickle
from lightningcast import utils
import glob
import copy

SEC_PER_DAY = 3600.0 * 24.0

# -----------------------------------------------------------------------------------------------------


def strip_accents(
    text,
):  # from https://gitlab.ssec.wisc.edu/volcat/volcat/-/blob/master/fire_incidents.py
    """
    Strip accent characters from a string
    """

    try:
        text = unicode(text, "utf-8")
    except (TypeError, NameError):  # unicode is a default on python 3
        pass

    text = unicodedata.normalize("NFD", text)
    text = text.encode("ascii", "ignore")
    text = text.decode("utf-8")

    return str(text)


# -----------------------------------------------------------------------------------------------------


def clean_string(
    string_in,
):  # from https://gitlab.ssec.wisc.edu/volcat/volcat/-/blob/master/fire_incidents.py
    """
    Make sure accents and other problematic characters are stripped from
    important strings
    """

    string_out = strip_accents(string_in)
    string_out = string_out.rstrip("\n")
    string_out = string_out.replace("&", "and")
    string_out = string_out.replace("/", "-")
    string_out = string_out.replace("'", "")
    string_out = string_out.replace(":", "")
    string_out = string_out.replace(";", "")
    string_out = string_out.replace(".", "")

    return string_out


# -----------------------------------------------------------------------------------------------------


def make_fire_points(
    point_data,
    epoch_target,
    epoch_target0,
    epoch_target1,
    update_threshold,
    min_incident_lon,
    max_incident_lon,
    min_incident_lat,
    max_incident_lat,
):
    """
    Make dictionary of data for fires we're interested in.
    epoch_target0 and epoch_target1 are bounding times for candidate fires'
    *inception times* (in seconds since 1970-01-01 00:00:00 UTC)
    The lats and lons are also bounding points for candidate fires' centroids.
    update_threshold for candidate fires (in days)
    """

    logging.info("Process started")
    # From wfigs_points function in https://gitlab.ssec.wisc.edu/volcat/volcat/-/blob/master/fire_incidents.py

    num_points = 0
    mpoints = []

    # FireDiscoveryDateTime
    # FireOutDateTime
    # 2021-01-28T21:45:05Z
    # ModifiedOnDateTime_dt
    # FinalFireReportApprovedDate
    # InitialLatitude
    # InitialLongitude
    # POOCounty
    # POOState
    # InitialResponseAcres
    # IncidentTypeCategory
    # IncidentTypeKind
    # IrwinID
    # GlobalID

    # Go through fire events in geojson
    if "features" in point_data.keys():

        logging.info("Found valid WFIGS point features")

        num_poly = 0

        for i in point_data["features"]:

            if i["properties"]["FireDiscoveryDateTime"] is not None:
                if str(i["properties"]["FireDiscoveryDateTime"]).isdigit():
                    epoch_start = int(
                        int(i["properties"]["FireDiscoveryDateTime"]) / 1000.0
                    )
                    t_start = datetime.datetime.utcfromtimestamp(epoch_start).strftime(
                        "%Y-%m-%dT%H:%M:%S"
                    )
                else:
                    t_start = np.datetime64(i["properties"]["FireDiscoveryDateTime"])
                    epoch_start = t_start.astype("datetime64[s]").astype("int32")

                if i["properties"]["ModifiedOnDateTime_dt"] is not None:
                    if str(i["properties"]["ModifiedOnDateTime_dt"]).isdigit():
                        epoch_last_update = int(
                            int(i["properties"]["ModifiedOnDateTime_dt"]) / 1000.0
                        )
                        t_last_update = datetime.datetime.utcfromtimestamp(
                            epoch_last_update
                        ).strftime("%Y-%m-%dT%H:%M:%S")
                    else:
                        t_last_update = np.datetime64(
                            i["properties"]["ModifiedOnDateTime_dt"]
                        )
                        epoch_last_update = t_last_update.astype(
                            "datetime64[s]"
                        ).astype("int32")
                    days_since_last_update = (
                        epoch_target - epoch_last_update
                    ) / SEC_PER_DAY
                else:
                    t_last_update = "NULL"
                    epoch_last_update = -999
                    days_since_last_update = -999

                # if(i['properties']['IncidentName'] == 'BRUSHY UNIT 6 RX'):
                #   print(days_since_last_update);sys.exit()

                if i["properties"]["FireOutDateTime"] is not None:
                    if str(i["properties"]["FireOutDateTime"]).isdigit():
                        epoch_end = int(
                            int(i["properties"]["FireOutDateTime"]) / 1000.0
                        )
                        t_end = datetime.datetime.utcfromtimestamp(epoch_end).strftime(
                            "%Y-%m-%dT%H:%M:%S"
                        )
                    else:
                        t_end = np.datetime64(i["properties"]["FireOutDateTime"])
                        epoch_end = t_end.astype("datetime64[s]").astype("int32")
                else:
                    t_end = "NULL"
                    epoch_end = -999

                if (
                    (epoch_end < 0 or epoch_end > epoch_target0)
                    and (epoch_start > epoch_target0)
                    and (epoch_start < epoch_target1)
                    and (days_since_last_update < update_threshold)
                ):
                    if (
                        i["geometry"]["coordinates"][0] > min_incident_lon
                        and i["geometry"]["coordinates"][0] < max_incident_lon
                        and i["geometry"]["coordinates"][1] > min_incident_lat
                        and i["geometry"]["coordinates"][1] < max_incident_lat
                    ):
                        num_points = num_points + 1

                        incident_name = clean_string(
                            i["properties"]["IncidentName"]
                        ).upper()
                        if (i['properties']['POOCounty']):
                            incident_county = clean_string(i['properties']['POOCounty']).upper()
                        else:
                            incident_county = 'UNKNOWN'

                        mpoints.append(
                            {
                                "t_start": t_start,
                                "epoch_start": epoch_start,
                                "t_end": t_end,
                                "epoch_end": epoch_end,
                                "t_last_update": t_last_update,
                                "epoch_last_update": epoch_last_update,
                                "incident_name": incident_name,
                                "incident_county": incident_county,
                                "incident_state": i["properties"]["POOState"],
                                #     'incident_cat': i['properties']['IncidentTypeCategory'],
                                #     'incident_kind': i['properties']['IncidentTypeKind'],
                                #     'IrwinID': i['properties']['IrwinID'],
                                #     'GlobalID': i['properties']['GlobalID'],
                                "incident_lon": i["geometry"]["coordinates"][0],
                                "incident_lat": i["geometry"]["coordinates"][1],
                            }
                        )

    logging.info("Process ended")
    return mpoints, num_points


# -----------------------------------------------------------------------------------------------------


def update_ltg_for_fires(
    mpoints,
    fire_events,
    target_dt,
    prob_lats=None,
    prob_lons=None,
    prob_data=None,
    L2_flashes=None,
):

    """
    Use the current fire events geoJSON data (mpoints)
    and the saved fire_events (from pickle with ltg time series data)
    and update fire_events with lighting probability and flash data.
    Remove any "old" or not-found incidents.
    """

    buff = fire_events["metadata"]["buff"]

    try:
        truncate_thresh = fire_events["metadata"][
            "truncate_thresh"
        ]  # in days. Threshsold to truncate ts data for a fire.
        delete_thresh = fire_events["metadata"][
            "update_thresh"
        ]  # in days. Threshold to delete ts data for a fire.
    except KeyError:
        logging.error(
            "No 'truncate_thresh' or 'delete_thresh' in fire_events['metadata']. Using defaults of 7 and 14 days."
        )
        truncate_thresh = 7  # days
        delete_thresh = 14  # days

    # Remove fires from fire_events are not in mpoints (usu. those that have not been updated recently).
    delete_IDs = []
    incident_names = [mm["incident_name"] for mm in mpoints]
    for ID in fire_events["fires"]:
        if ID not in incident_names:
            logging.info(
                f"{ID} in fire_events is not in latest incident geoJSON. Removing."
            )
            delete_IDs.append(ID)
    for ID in delete_IDs:
        del fire_events["fires"][ID]

    # Go through list of current fires from fire_event_geojson.
    # Update or create ts data for these events. Delete old ones.
    delete_IDs = []
    for i in mpoints:
        ID = i["incident_name"]
        if ID not in fire_events["fires"]:  # new incident
            fire_events["fires"][ID] = {
                "pltg60min": [],  # LightningCast probability at the fire event point
                "max_pltg60min_20km": [],  # Maximum LightningCast probability within 20km of the fire event point
                "glm_flashes_10km": [],  # Number of GLM flashes within 10km of the fire event point
                "glm_flashes_20km": [],  # Number of GLM flashes within 20km of the fire event point
                # 'fed':[],                #Flash-extent density at the fire event point
                # 'max_fed_20km':[],       #Max flash-extent density within 20km of the fire event point
                "datetimes": [],  # <datetime.dateime> objects
                "RElink": "foo",
            }  # URL to see zoom-in of LC, GLM, and ABI data around fire event
            if i["epoch_last_update"] > 0:
                update_dt = datetime.datetime.utcfromtimestamp(i["epoch_last_update"])
                td_last_update = target_dt - update_dt  # "td" is for "time delta"
                if td_last_update.days >= truncate_thresh:
                    # Truncate time series to latest 24 hours of data.
                    logging.info(
                        f"Last update for {ID} was >= {truncate_thresh} days ago. Truncating time series to latest 1 day of data."
                    )
                    # Search datetimes backwards. Break when we are older than 24 hours.
                    for dd, dt in enumerate(
                        reversed(fire_events["fires"][ID]["datetimes"])
                    ):
                        if (target_dt - dt).days >= 1:
                            cutoff_idx = (
                                dd * (-1) - 1
                            )  # make negative and subtract 1 since -1 is the first index when counting backwards.
                            for key in fire_events["fires"][ID]:
                                fire_events["fires"][ID][key] = fire_events["fires"][
                                    ID
                                ][key][cutoff_idx:]
                            break
            else:
                # No end or update times, so use the start time of the fire
                start_dt = datetime.datetime.utcfromtimestamp(i["epoch_start"])
                td_start = target_dt - start_dt  # "td" is for "time delta"
                if td_start.days >= delete_thresh:
                    logging.info(
                        f"{ID} began >= {delete_thresh} days ago. {ID} will be deleted from fire_events."
                    )
                    delete_IDs.append(ID)

        # Update the valid fire events with lightning probability and flash data
        if ID not in delete_IDs:
            # Find closest lat and lon, and numpy indices to get prob data.
            flat = i["incident_lat"]
            flon = i["incident_lon"]
            fire_events["fires"][ID]["datetimes"].append(target_dt)
            if (
                prob_data is not None
            ):  # i.e., if LightningCast probabilities were provided
                latidx, _ = utils.find_nearest(prob_lats, flat)
                lonidx, _ = utils.find_nearest(prob_lons, flon)

                # Append new lightning probabilities. Assumes 0.01 = 1km
                try:
                    fire_events["fires"][ID]["pltg60min"].append(
                        prob_data[latidx, lonidx]
                    )
                    fire_events["fires"][ID]["max_pltg60min_20km"].append(
                        np.max(
                            prob_data[
                                latidx - buff : latidx + buff + 1,
                                lonidx - buff : lonidx + buff + 1,
                            ]
                        )
                    )
                except ValueError as err:
                    logging.warning(
                        f"({np.round(flat,3)},{np.round(flon,3)}) is off of the LightningCast grid. Filling probabilities with missing values."
                    )
                    fire_events["fires"][ID]["pltg60min"].append(-1)  # missing val
                    fire_events["fires"][ID]["max_pltg60min_20km"].append(
                        -1
                    )  # missing val

            if (
                L2_flashes is not None
            ):  # i.e., if using GLM L2 flashes, get count of nearby flashes (points). Assumes 0.01 = 1km
                if len(L2_flashes["lats"]) > 0:
                    fire_events["fires"][ID]["glm_flashes_10km"].append(
                        np.count_nonzero(
                            (L2_flashes["lats"] >= (flat - 0.1))
                            & (L2_flashes["lats"] <= (flat + 0.1))
                            & (L2_flashes["lons"] >= (flon - 0.1))
                            & (L2_flashes["lons"] <= (flon + 0.1))
                        )
                    )
                    fire_events["fires"][ID]["glm_flashes_20km"].append(
                        np.count_nonzero(
                            (L2_flashes["lats"] >= (flat - 0.2))
                            & (L2_flashes["lats"] <= (flat + 0.2))
                            & (L2_flashes["lons"] >= (flon - 0.2))
                            & (L2_flashes["lons"] <= (flon + 0.2))
                        )
                    )

            # Create RElink
            try:
                sat = fire_events["metadata"]["satellite"]
                orbslot = fire_events["metadata"]["orbital_slot"]
                sector = fire_events["metadata"]["sector"]
            except KeyError:
                logging.warning(
                    "One or more necessary keys missing in metadata. RElink can't be created."
                )
            else:
                if sector == "RadC":
                    sec = "CONUS"
                elif sector == "RadM1":
                    sec = "MESO1"
                elif sector == "RadM2":
                    sec = "MESO2"
                elif sector == "RadF":
                    sec = "FD"
                else:
                    logging.error(
                        f"Sector {sec} not supported. RE link will not be fully functional."
                    )
                RElink = (
                    f"https://realearth.ssec.wisc.edu/?products={sat}-ABI-{sec}-BAND13-GRAD.100,{sat}-ABI-{sec}-BAND02.66,"
                    + f"{orbslot}GLMFEDRadC.100,PLTG{orbslot}{sector}.100&center={flat},{flon}&zoom=8&width=1368&height=747"
                    + f"&basemap=terrain&labels=lines&view=leaflet&timeproduct=PLTG{orbslot}{sector}&timespan=-1h&animationspeed=50"
                )
                fire_events["fires"][ID]["RElink"] = RElink

    # Delete "old" fire events
    for ID in delete_IDs:
        try:
            del fire_events["fires"][ID]
        except KeyError:
            pass


# -----------------------------------------------------------------------------------------------------


def get_L2_flashes(glm_L2_patt, start_dt, end_dt, plax_corr_hgt=0):
    """
    Uses the glm_L2_patt and start_dt and end_dt to get GLM L2 flashes.
    The program will gather flash lat/lon data from startdt to end_dt, iterating by 1 min.
    Returns a dictionary of GLM-deteced flash lats/lons between start_dt and end_dt.
    Will perform parallax correction of flashes using the plax_corr_hgt (in m AGL).
    (I.e., the lats/lons of the flashes will be altered assuming a constant parallax cloud height).
    """

    logging.info("Process started.")

    assert start_dt <= end_dt
    satlon = -999

    L2_flashes = {"lats": [], "lons": []}

    dt = start_dt
    while dt <= end_dt:
        files = np.sort(glob.glob(dt.strftime(glm_L2_patt)))
        for f in files:  # is usu. 3 files; every 20 sec for 1 min
            logging.info(f"Reading {os.path.basename(f)}")
            nc = netCDF4.Dataset(f, "r")
            L2_flashes["lats"] += list(nc.variables["flash_lat"][:])
            L2_flashes["lons"] += list(nc.variables["flash_lon"][:])
            if satlon == -999:  # needed for plax correction
                satlon = nc.variables["nominal_satellite_subpoint_lon"][:]
            nc.close()
        dt = dt + datetime.timedelta(minutes=1)

    if len(L2_flashes["lats"]) > 0:
        # Parallax correct lats and lons
        if plax_corr_hgt > 0:
            logging.info(
                f"Parallax-correcting flash lats and lons using constant parallax height={plax_corr_hgt}m"
            )
            L2_flashes["lats"], L2_flashes["lons"] = utils.plax(
                lats=np.array(L2_flashes["lats"]),
                lons=np.array(L2_flashes["lons"]),
                satlon=satlon,
                elevation=plax_corr_hgt,
            )
    else:
        logging.warning("No GLM flashes found!")

    logging.info("Process ended.")
    return L2_flashes


# -----------------------------------------------------------------------------------------------------


def blank_fire_dict(target_dt, update_threshold):  # some stuff hard-coded for now

    fire_events = {
        "fires": {},
        "metadata": {
            "valid_time": target_dt,
            "truncate_thresh": 7,
            "update_thresh": update_threshold,
        },
    }
    return fire_events


# -----------------------------------------------------------------------------------------------------


def fire_event_timeseries(
    fire_event_geojson="/apollo/grain/common/NIFC/Current-Fire-Location.geojson",
    outdir=os.environ["PWD"],
    fire_event_pickle=None,
    realtime=False,
    prob_netcdf=None,
    glm_netcdf=None,
    glm_L2_patt=None,
    ll_bbox=[-999, 999, -999, 999],
    target_dt=None,
    min_dt=None,
    max_dt=None,
    update_threshold=14,  # days
):
    logging.info("Process started.")

    if (
        glm_L2_patt
    ):  # Default is to use glm_L2_patt. If user turns it off by making it "None", set to None.
        if "none" in glm_L2_patt.lower():
            glm_L2_patt = None
            logging.info("Setting glm_L2_patt to None.")

    min_incident_lon, max_incident_lon, min_incident_lat, max_incident_lat = ll_bbox
    if not (target_dt):
        # Default to time of prob_netcdf. If that doesn't exist, use datetime.now()
        if prob_netcdf:
            # LightningCast_GOES-16_RadC_20221129-200617.nc
            timestr = os.path.basename(prob_netcdf).split("_")[-1]
            target_dt = datetime.datetime.strptime(timestr, "%Y%m%d-%H%M%S.nc")
        else:
            target_dt = datetime.datetime.now()
    if not (min_dt):
        min_dt = datetime.datetime(1970, 1, 1, 0, 0)
    if not (max_dt):
        max_dt = datetime.datetime(3099, 12, 31, 0, 0)

    # Set epoch time info. All in seconds since 01/01/1970 00:00 UTC
    assert isinstance(target_dt, datetime.datetime) or isinstance(target_dt, str)
    assert isinstance(min_dt, datetime.datetime) or isinstance(min_dt, str)
    assert isinstance(max_dt, datetime.datetime) or isinstance(max_dt, str)
    if isinstance(target_dt, str):
        target_dt = datetime.datetime.strptime(target_dt, "%Y%m%d-%H%M")
    if isinstance(min_dt, str):
        min_dt = datetime.datetime.strptime(min_dt, "%Y%m%d-%H%M")
    if isinstance(max_dt, str):
        max_dt = datetime.datetime.strptime(max_dt, "%Y%m%d-%H%M")
    epoch_target = int(
        (target_dt - datetime.datetime(1970, 1, 1, 0, 0)).total_seconds()
    )
    epoch_target0 = int((min_dt - datetime.datetime(1970, 1, 1, 0, 0)).total_seconds())
    epoch_target1 = int((max_dt - datetime.datetime(1970, 1, 1, 0, 0)).total_seconds())

    # Read the fire_event_pickle. This is a dict containing time series to be appended for fire events (or add new events).
    if fire_event_pickle:
        try:
            fire_events = pickle.load(open(fire_event_pickle, "rb"))
        except FileNotFoundError:
            logging.warning(
                f"Couldn't find/open {fire_event_pickle}. Using blank dictionary."
            )
            fire_events = blank_fire_dict(target_dt, update_threshold)
    else:
        fire_events = blank_fire_dict(target_dt, update_threshold)
    fire_events["metadata"]["valid_time"] = target_dt

    # Get gridded ltg prob data
    if prob_netcdf:
        probvar = "PLTG_60min_plax"  # Assume we want parallax-corrected field. This is on an equal-lat-lon grid.
        prob_ds = {probvar: {"atts": {}}}
        global_atts = {
            "NWlat": -1,
            "NWlon": -1,
            "dx": -1,
            "dy": -1,
            "satellite": -1,
            "orbital_slot": -1,
            "sector": -1,
        }
        utils.read_netcdf(prob_netcdf, datasets=prob_ds, global_atts=global_atts)
        prob_data = prob_ds[probvar]["data"]
        nlat, nlon = prob_data.shape
        slat = global_atts["NWlat"] - nlat * global_atts["dy"]
        prob_lats = np.linspace(global_atts["NWlat"], slat, nlat)
        elon = global_atts["NWlon"] + nlon * global_atts["dx"]
        prob_lons = np.linspace(global_atts["NWlon"], elon, nlon)
        # Get buffer in pixels. For "p(ltg) within 20km, e.g. 20 km = 0.2
        buff = int(np.round(0.2 / global_atts["dy"]))  # assumes dy=dx
        # Update satellite/sector metadata
        fire_events["metadata"]["satellite"] = global_atts["satellite"].replace(
            "OES-", ""
        )
        fire_events["metadata"]["orbital_slot"] = global_atts["orbital_slot"].replace(
            "-", ""
        )
        fire_events["metadata"]["sector"] = global_atts["sector"]
        fire_events["metadata"]["buff"] = buff
        outfile = f"{outdir}/Current_Fire_Timeseries_{global_atts['satellite']}_{global_atts['sector']}.pkl"
    else:
        prob_data = prob_lons = prob_lats = buff = None
        outfile = f"{outdir}/Current_Fire_Timeseries.pkl"

    # Get gridded GLM data
    if glm_netcdf:
        logging.critical("Not implemented yet.")
        sys.exit(1)

    # Get GLM flashes from L2 files
    if glm_L2_patt:
        # L2_flashes is a dict:{'lats':[], 'lons':[]}
        fire_events["metadata"]["GLM_flash_point_time_window"] = "last 5 min"
        L2_flashes = get_L2_flashes(
            glm_L2_patt,
            start_dt=target_dt - datetime.timedelta(minutes=5),
            end_dt=target_dt,
            plax_corr_hgt=9000,
        )
    else:
        L2_flashes = None

    # Read the fire_event_geojson
    logging.info(f"Reading {fire_event_geojson}")
    f = open(fire_event_geojson)
    point_data = json.load(f)
    f.close()

    # Get list of current fire events/dicts from geoJSON (mpoints)
    mpoints, num_points = make_fire_points(
        point_data,
        epoch_target,
        epoch_target0,
        epoch_target1,
        update_threshold,
        min_incident_lon,
        max_incident_lon,
        min_incident_lat,
        max_incident_lat,
    )

    # Update fire events with lightning/prob data.
    # fire_events gets returned modified.
    update_ltg_for_fires(
        mpoints,
        fire_events,
        target_dt,
        prob_lats=prob_lats,
        prob_lons=prob_lons,
        prob_data=prob_data,
        L2_flashes=L2_flashes,
    )

    if realtime:
        # If realtime, we want to return a dictionary of fires' properties with one timestep of data.
        # Go through each property of each fire event and take 0th element if it's a list.
        tmp_fire_events = copy.deepcopy(fire_events)
        for ID, event_data in tmp_fire_events[
            "fires"
        ].items():  # fire_events['fires'][ID] = {'pltg':[...], 'glm_flashes':[...]}
            for key, prop in event_data.items():
                if isinstance(
                    prop, list
                ):  # If property is a list, take the 0th element (should only be one)
                    fire_events["fires"][ID][key] = prop[0]
            # Translate the satellite info from metadata to properties for each event ID
            fire_events["fires"][ID]["satellite"] = fire_events["metadata"]["satellite"]
            fire_events["fires"][ID]["orbital_slot"] = fire_events["metadata"][
                "orbital_slot"
            ]
            fire_events["fires"][ID]["sector"] = fire_events["metadata"]["sector"]

        logging.info("Returning fire_events['fires']")
        return fire_events["fires"]

    else:
        # Output the file
        utils.mkdir_p(outdir)
        addon = "Overwrote file." if (os.path.isfile(outfile)) else ""
        pickle.dump(fire_events, open(outfile, "wb"))
        logging.info(f"Saved {outfile}. {addon}")

        logging.info("Process ended.")


# -----------------------------------------------------------------------------------------------------

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Takes a geoJSON FeatureList of fire events (points), P(LTG) grid, \
                                                GLM flash points/grid, and creates or updates time series of P(LTG) and \
                                                GLM flashes at or near those events."
    )
    parser.add_argument(
        "--fire_event_geojson",
        "-f",
        help="A geoJSON FeatureList of fire events. \
                      Default=/apollo/grain/common/NIFC/Current-Fire-Location.geojson.",
        default=["/apollo/grain/common/NIFC/Current-Fire-Location.geojson"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--fire_event_pickle",
        "-p",
        help="A pickle file containing time series of lightning data for fire incidents. \
                      This contains the time series to be appended. Dict is written to disk rather than returned. Cannot be used with --realtime. \
                      If not --realtime and --fire_event_pickle is not supplied, will use a blank dict.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-r",
        "--realtime",
        help="Realtime mode. The main function will return a dict with data for one timestep \
                      for each fire event. Cannot be used with --fire_event_pickle.",
        action="store_true",
    )
    parser.add_argument(
        "--outdir",
        "-o",
        help="Output directory. Default = $PWD.",
        default=[os.environ["PWD"]],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--prob_netcdf",
        "-pf",
        help="LightningCast probability netcdf. Assumes coverage of all fire events.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--glm_netcdf",
        "-gf",
        help="GLM gridded netcdf. Assumes coverage of all fire events.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--glm_L2_patt",
        "-gp",
        help="GLM L2 pattern. Assumes coverage of all fire events. \
                      Default = '/arcdata/goes/grb/goes16/%%Y/%%Y_%%m_%%d_%%j/glm/L2/LCFA/*_s%%Y%%j%%H%%M*'. \
                      Set to 'None' if you don't want to get glm flash points.",
        default=["/arcdata/goes/grb/goes16/%Y/%Y_%m_%d_%j/glm/L2/LCFA/*_s%Y%j%H%M*"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--target_dt",
        "-t",
        help="Usually the satellite scan time. E.g., 20221123-1241. \
                      Default=<prob_netcdf dt> (if available), else default=<NOW>.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--min_dt",
        "-mt",
        help="Minimum dt cutoff for events. E.g., 20221123-1241. Default is no cutoff.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--max_dt",
        "-xt",
        help="Maximum dt cutoff for events. E.g., 20221123-1241. Default is no cutoff.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "--latlon_bbox",
        "-bb",
        help="minlon maxlon minlat maxlat. Default is -999 999 -999 999.",
        default=[-999, 999, -999, 999],
        type=int,
        nargs=4,
    )
    parser.add_argument(
        "--update_threshold",
        "-ut",
        help="Threshold in days such that any fire event that hasn't been updated \
                      within that threshold will get purged. Default=14.",
        default=[14],
        type=int,
        nargs=1,
    )

    args = parser.parse_args()

    logDict = logging_def(logfile=None)
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)

    if args.target_dt[0]:
        target_dt = datetime.datetime.strptime(args.target_dt[0], "%Y%m%d-%H%M")
    else:
        # Default to time of prob_netcdf. If that doesn't exist, use datetime.now()
        if args.prob_netcdf[0]:
            # LightningCast_GOES-16_RadC_20221129-200617.nc
            timestr = os.path.basename(args.prob_netcdf[0]).split("_")[-1]
            target_dt = datetime.datetime.strptime(timestr, "%Y%m%d-%H%M%S.nc")
        else:
            target_dt = datetime.datetime.now()

    min_dt = (
        datetime.datetime.strptime(args.min_dt[0], "%Y%m%d-%H%M")
        if (args.min_dt[0])
        else datetime.datetime(1970, 1, 1, 0, 0)
    )
    max_dt = (
        datetime.datetime.strptime(args.max_dt[0], "%Y%m%d-%H%M")
        if (args.max_dt[0])
        else datetime.datetime(3099, 12, 31, 0, 0)
    )

    fire_event_timeseries(
        fire_event_geojson=args.fire_event_geojson[0],
        fire_event_pickle=args.fire_event_pickle[0],
        realtime=args.realtime,
        outdir=args.outdir[0],
        prob_netcdf=args.prob_netcdf[0],
        glm_netcdf=args.glm_netcdf[0],
        glm_L2_patt=args.glm_L2_patt[0],
        target_dt=target_dt,
        min_dt=min_dt,
        max_dt=max_dt,
        ll_bbox=args.latlon_bbox,
        update_threshold=args.update_threshold[0],
    )
