import numpy as np
import sys, os

import inspect, sys

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
target_dir = os.path.dirname(os.path.dirname(current_dir))
sys.path.insert(0, target_dir)

from netCDF4 import Dataset
from datetime import datetime, timedelta
import time
import argparse
from glob import glob
import pyresample as pr
from lightningcast import utils
from pyproj import Proj
import pyproj
import subprocess
import matplotlib.pyplot as plt

# from mpl_toolkits.basemap import Basemap
import collections
import numpy.ma as ma
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import traceback
from scipy.ndimage import zoom
import pickle
import rasterio
import rasterio.transform

#####################################################################################################################################
def main():
    parse_desc = """Take gridded GLM fields from Eric Bruning's glmtools and aggregate them. Optionally, make a geotiff.
                  Note that this should be used on 1-min aggregations only."""

    parser = argparse.ArgumentParser(description=parse_desc)
    parser.add_argument(
        "lastdatetime",
        help="Datetime of latest file to include in aggregation. E.g., 20181621515 -- yyyyjjjhhmm",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "datadir",
        help="directory of netcdf files. E.g., /data/GLM . Subdirs assumed (e.g.): /data/GLM/2018/Jun/25/",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-a",
        "--accumPeriod",
        help="Number of minutes to accumulate. Default = 5.",
        default=[5],
        type=int,
        nargs=1,
    )
    parser.add_argument(
        "-o", "--outdir", help="Output directory. Default = ${PWD}", type=str, nargs=1
    )
    parser.add_argument(
        "-gt",
        "--geotiff",
        help="Set this flag to create geotiff products",
        action="store_true",
    )
    parser.add_argument(
        "-gzip",
        "--gzip_prods",
        help="gzip the GLM grids? Default = False",
        action="store_true",
    )
    parser.add_argument(
        "-l", "--logfile", help="print to this logfile", default=[None], nargs=1
    )
    parser.add_argument(
        "-ru",
        "--re_upload",
        help="Full path to re_upload executable. Will execute after GTiff is written.",
        default=[None],
        type=str,
        nargs=1,
    )

    args = parser.parse_args()

    if not args.datadir:
        datadir = "None"
    else:
        datadir = args.datadir[0]
    if not args.outdir:
        outdir = os.environ["PWD"] + "/"
    else:
        outdir = args.outdir[0]

    logDict = logging_def(logfile=args.logfile[0])
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    status = aggregate_glm_grids(
        args.lastdatetime[0],
        args.datadir[0],
        accumPeriod=args.accumPeriod[0],
        outdir=outdir,
        geotiff=args.geotiff,
        re_upload=args.re_upload[0],
        gzip=args.gzip_prods,
    )



#################################################################################################################################################
def aggregate_glm_grids(
    lastdatetime,
    datadir,
    accumPeriod=5,
    outdir=os.environ["PWD"] + "/",
    geotiff=False,
    re_upload=None,
    gzip=False,
):

    if datadir[-1] != "/":
        datadir += "/"

    logging.info("Process started")

    try:
        last_dt = datetime.strptime(lastdatetime, "%Y%j%H%M")
    except ValueError as err:
        logging.error(traceback.format_exc())
        return False

    # aggregate data for the accumPeriod
    accumMins = 0
    for tt in range(accumPeriod):
        curr_dt = last_dt - timedelta(minutes=tt)
        if tt == 0:
            ts_dt = curr_dt

        filepattern = datadir + curr_dt.strftime("%Y/%b/%d/OR_GLM-L2-*s%Y%j%H%M*.nc*")

        ncfile = np.sort(glob(filepattern))
        if len(ncfile):
            gzip = False
            logging.info(f"received {ncfile[0]}")
            if (
                tt == 0
            ):  # get end datetime of first file in accumulation, since we're working back in time.
                ts_dt = datetime.strptime(
                    os.path.basename(ncfile[0]).split("_e")[1].split("_")[0][0:13],
                    "%Y%j%H%M%S",
                )
            accumMins += 1
            ncfile = ncfile[0]
            try:
                nc = Dataset(ncfile, "r")
            except (OSError, IOError) as err:
                ncfile_gzipped = ncfile
                ncfile = ncfile_gzipped[0 : len(ncfile_gzipped) - 3]
                s = subprocess.call(["gunzip", "-f", ncfile_gzipped])
                gzip = True
                nc = Dataset(ncfile, "r")

            flash_extent_density_TMP = nc.variables["flash_extent_density"][:]

            average_area_perflash = nc.variables["average_flash_area"][:]
            average_area_TMP = np.multiply(
                average_area_perflash, flash_extent_density_TMP
            )

            group_extent_density_TMP = nc.variables["group_extent_density"][:]

            average_area_pergroup = nc.variables["average_group_area"][:]
            average_garea_TMP = np.multiply(
                average_area_pergroup, group_extent_density_TMP
            )

            flash_centroid_density_TMP = nc.variables["flash_centroid_density"][:]

            total_energy_TMP = (
                nc.variables["total_energy"][:] * 1e6
            )  # in nJ...put into fJ

            group_centroid_density_TMP = nc.variables["group_centroid_density"][:]

            # first see if the vars exist (i.e., it's the first iteration if they do not)
            try:
                flash_extent_density
            except NameError:
                # if they do not exist, copy fields
                flash_extent_density = np.copy(flash_extent_density_TMP)
                total_energy = np.copy(total_energy_TMP)
                average_area_AGG = np.copy(average_area_TMP)
                average_garea_AGG = np.copy(average_garea_TMP)
                flash_centroid_density = np.copy(flash_centroid_density_TMP)
                group_extent_density = np.copy(group_extent_density_TMP)
                group_centroid_density = np.copy(group_centroid_density_TMP)
                ydim = nc.dimensions["y"].size
                xdim = nc.dimensions["x"].size
                min_flash_area_grids = np.zeros((accumPeriod, ydim, xdim))
                min_flash_area_grids[tt] = nc.variables["minimum_flash_area"][:]
                # and grab x and y
                x = np.array(nc["x"][:])
                y = np.array(nc["y"][:])
                gip = nc["goes_imager_projection"][:]
            else:
                # if they do exist, just aggregate fields
                try:
                    flash_extent_density += flash_extent_density_TMP
                    total_energy += total_energy_TMP
                    average_area_AGG += average_area_TMP
                    average_garea_AGG += average_garea_TMP
                    flash_centroid_density += flash_centroid_density_TMP
                    group_extent_density += group_extent_density_TMP
                    group_centroid_density += group_centroid_density_TMP
                    min_flash_area_grids[tt] = nc.variables["minimum_flash_area"][:]
                except ValueError as err:
                    logging.error(traceback.format_exc())
                    nc.close()
                    return False

            # get attributes
            try:
                vatts
                gatts
            except NameError:
                vatts = collections.OrderedDict()
                vatts["x"] = {}
                vatts["y"] = {}
                vatts["goes_imager_projection"] = {}
                vatts["flash_extent_density"] = {}
                vatts["flash_centroid_density"] = {}
                vatts["total_energy"] = {}
                vatts["average_flash_area"] = {}
                vatts["group_extent_density"] = {}
                vatts["group_centroid_density"] = {}
                vatts["average_group_area"] = {}
                vatts["minimum_flash_area"] = {}
                for var in vatts:
                    for att in nc.variables[var].ncattrs():
                        vatts[var][att] = nc.variables[var].getncattr(att)
                vatts["total_energy"]["units"] = "fJ"
                vatts["total_energy"]["scale_factor"] = 1
                vatts["total_energy"]["add_offset"] = 0
                vatts["average_flash_area"]["units"] = "km^2"
                vatts["average_group_area"]["units"] = "km^2"

                # global atts
                gatts = collections.OrderedDict()
                gatts["spatial_resolution"] = getattr(nc, "spatial_resolution")
                gatts["subsatellite_longitude"] = vatts["goes_imager_projection"][
                    "longitude_of_projection_origin"
                ]
                gatts["orbital_slot"] = getattr(nc, "orbital_slot")
                gatts["platform_ID"] = getattr(nc, "platform_ID")
                gatts["scene_id"] = gatts["orbital_slot"][5] + getattr(
                    nc, "scene_id"
                )  # gatts['orbital_slot'][5] is "E" or "W"
                gatts["time_coverage_end"] = getattr(nc, "time_coverage_end")

            # do every time
            gatts["time_coverage_start"] = getattr(nc, "time_coverage_start")

            # close ncfile
            nc.close()
            if gzip:
                sss = subprocess.Popen(
                    ["gzip", "-f", ncfile]
                )  # ==will run in background

    if accumMins == 0:
        return False
    gatts["aggregation"] = str(accumMins) + " min"
    gatts["dataset_name"] = "GLM_gridded_data"

    for var in [
        "flash_extent_density",
        "flash_centroid_density",
        "group_extent_density",
        "group_centroid_density",
    ]:
        attparts = vatts[var]["units"].split(" ")
        attparts[-2] = str(accumMins)
        vatts[var]["units"] = (" ").join(attparts)

    # Make final min area
    fillv = 999999
    min_flash_area_grids[min_flash_area_grids == 0] = fillv
    minimum_flash_area = np.min(min_flash_area_grids, axis=0)
    minimum_flash_area[minimum_flash_area == fillv] = 0

    # Make average areas.
    # These steps are necessary so we don't divide by 0 or some small fraction
    average_flash_area = np.zeros((np.shape(average_area_AGG)))
    good_ind = np.where(flash_extent_density >= 1)
    average_flash_area[good_ind] = (
        average_area_AGG[good_ind] / flash_extent_density[good_ind]
    )
    average_flash_area = average_flash_area.astype("f4")

    average_group_area = np.zeros((np.shape(average_garea_AGG)))
    good_ind = np.where(group_extent_density >= 1)
    average_group_area[good_ind] = (
        average_garea_AGG[good_ind] / group_extent_density[good_ind]
    )
    average_group_area = average_group_area.astype("f4")

    ny, nx = np.shape(total_energy)

    # write out data into one file
    # first pack the data using scale_factor and add_offset
    xpacked = ((x - vatts["x"]["add_offset"]) / vatts["x"]["scale_factor"]).astype(
        np.int16
    )
    ypacked = ((y - vatts["y"]["add_offset"]) / vatts["y"]["scale_factor"]).astype(
        np.int16
    )
    FEDpacked = (
        (flash_extent_density - vatts["flash_extent_density"]["add_offset"])
        / vatts["flash_extent_density"]["scale_factor"]
    ).astype(np.int16)
    FCDpacked = (
        (flash_centroid_density - vatts["flash_centroid_density"]["add_offset"])
        / vatts["flash_centroid_density"]["scale_factor"]
    ).astype(np.int16)
    TOEpacked = (
        (total_energy - vatts["total_energy"]["add_offset"])
        / vatts["total_energy"]["scale_factor"]
    ).astype(np.int16)
    AFApacked = (
        (average_flash_area - vatts["average_flash_area"]["add_offset"])
        / vatts["average_flash_area"]["scale_factor"]
    ).astype(np.int16)
    GEDpacked = (
        (group_extent_density - vatts["group_extent_density"]["add_offset"])
        / vatts["group_extent_density"]["scale_factor"]
    ).astype(np.int16)
    GCDpacked = (
        (group_centroid_density - vatts["group_centroid_density"]["add_offset"])
        / vatts["group_centroid_density"]["scale_factor"]
    ).astype(np.int16)
    AGApacked = (
        (average_group_area - vatts["average_group_area"]["add_offset"])
        / vatts["average_group_area"]["scale_factor"]
    ).astype(np.int16)
    MFApacked = (
        (minimum_flash_area - vatts["minimum_flash_area"]["add_offset"])
        / vatts["minimum_flash_area"]["scale_factor"]
    ).astype(np.int16)

    dims = collections.OrderedDict()
    dims["y"] = ny
    dims["x"] = nx
    datasets = collections.OrderedDict()
    datasets["x"] = {"data": xpacked, "dims": ("x"), "atts": vatts["x"]}
    datasets["y"] = {"data": ypacked, "dims": ("y"), "atts": vatts["y"]}
    datasets["goes_imager_projection"] = {
        "data": gip,
        "dims": (),
        "atts": vatts["goes_imager_projection"],
    }
    datasets["flash_extent_density"] = {
        "data": FEDpacked,
        "dims": ("y", "x"),
        "atts": vatts["flash_extent_density"],
    }
    datasets["flash_centroid_density"] = {
        "data": FCDpacked,
        "dims": ("y", "x"),
        "atts": vatts["flash_centroid_density"],
    }
    datasets["total_energy"] = {
        "data": TOEpacked,
        "dims": ("y", "x"),
        "atts": vatts["total_energy"],
    }
    datasets["average_flash_area"] = {
        "data": AFApacked,
        "dims": ("y", "x"),
        "atts": vatts["average_flash_area"],
    }
    datasets["group_extent_density"] = {
        "data": GEDpacked,
        "dims": ("y", "x"),
        "atts": vatts["group_extent_density"],
    }
    datasets["group_centroid_density"] = {
        "data": GCDpacked,
        "dims": ("y", "x"),
        "atts": vatts["group_centroid_density"],
    }
    datasets["average_group_area"] = {
        "data": AGApacked,
        "dims": ("y", "x"),
        "atts": vatts["average_group_area"],
    }
    datasets["minimum_flash_area"] = {
        "data": MFApacked,
        "dims": ("y", "x"),
        "atts": vatts["minimum_flash_area"],
    }

    grid_outdir = outdir + "/agg/"

    timestamp = ts_dt.strftime("%Y%m%d-%H%M%S")
    epoch_secs = int((ts_dt - datetime.utcfromtimestamp(0)).total_seconds())
    outfile = grid_outdir + timestamp + ".netcdf"

    status = utils.write_netcdf(
        outfile, datasets, dims, atts=gatts, wait=False, gzip=gzip, nc4=True
    )

    ######################################################
    ############# Now make geotiffs data #################
    ######################################################

    if geotiff:
        gtiff_outdir = f"{outdir}/gtiff/"
        utils.mkdir_p(gtiff_outdir)
        logging.info("Making geotiff...")
        cmap = plt.get_cmap("jet")
        colors = cmap(np.arange(0, cmap.N))  # colors.shape = (256,4) rgbas
        stride = 1

        # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
        projx = x * vatts["goes_imager_projection"]["perspective_point_height"]
        projy = y * vatts["goes_imager_projection"]["perspective_point_height"]
        proj_def_string = (
            f"+proj=geos +sweep=x +lon_0={vatts['goes_imager_projection']['longitude_of_projection_origin']} "
            + f"+h={vatts['goes_imager_projection']['perspective_point_height']} "
            + "+x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs"
        )
        xmin = projx[0]
        ymax = projy[0]
        xsize = projx[stride] - projx[0]
        ysize = projy[0] - projy[stride]
        transform = rasterio.transform.from_origin(xmin, ymax, xsize, ysize)
        fed = flash_extent_density[0::stride, 0::stride]
        # scale the grid to 0-255, using power normalization
        gamma = 0.5
        fed = fed ** (gamma)
        maxval = 128 ** (gamma)
        minval = 0.0 ** (
            gamma
        )  # scaling from 0 to 128 flashes #Should minval be 0 or 0.01?
        fed[fed < minval] = minval
        fed[fed > maxval] = maxval
        fed = ((fed - minval) * (1 / (maxval - minval) * 255)).astype("uint8")
        # fill tiffgrid with r,g,b triplets (0 to 1)
        ny, nx = np.shape(fed)
        tiffgrid = np.zeros((4, ny, nx), dtype=np.float32)  # rgba

        goodind = np.where(fed > 0)
        if len(goodind[0]):
            for ij in zip(goodind[0], goodind[1]):
                tiffgrid[:, ij[0], ij[1]] = colors[fed[ij]]
        # scale tiffgrid from 0 to 1 , to 0 to 255
        tiffgrid = ((tiffgrid - 0) * (1 / (1 - 0) * 255)).astype("uint8")

        slot = gatts["orbital_slot"].replace("-", "")

        if "ussamoa" in gtiff_outdir:
            outtif = ts_dt.strftime(
                f"{gtiff_outdir}/{slot}GLMFED-USSAMOA_%Y%m%d_%H%M00.force.tif"
            )
        else:
            outtif = ts_dt.strftime(
                f"{gtiff_outdir}/{slot}GLMFEDRadC_%Y%m%d_%H%M00.force.tif"
            )
        # count = number of bands
        new_dataset = rasterio.open(
            outtif,
            "w",
            driver="GTiff",
            height=ny,
            width=nx,
            count=4,
            dtype=str(tiffgrid.dtype),
            crs=proj_def_string,
            transform=transform,
        )
        new_dataset.write(tiffgrid)  # 3 is number of bands
        new_dataset.close()

        # from rasterio.plot import show
        # img = rasterio.open(outtif)
        # show(img)

        # re_upload is either None or a string to full path of the re_upload script
        if re_upload:
            try:
                s = subprocess.call([re_upload, '-o', outtif], timeout=10)
            except subprocess.TimeoutExpired as err:
                logging.error(str(err))
            else:
                logging.info(f"Completed {outtif}")
                os.remove(outtif)  # clean up

        return True

if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(str(e))