import numpy as np
import glob
import os, sys
from subprocess import call, Popen, PIPE
import argparse
from datetime import datetime, timedelta
import time
from netCDF4 import Dataset
from lightningcast.logging_def import *
from lightningcast import utils
from collections import OrderedDict
import traceback
from shapely.geometry import MultiPoint
from lightningcast.glm_objs_json import glm_objs_json
from sklearn.cluster import DBSCAN
from sklearn import metrics
from sklearn.preprocessing import StandardScaler
import shutil
import copy

eternity = "30991231-000000"

# =====================================================================================================================================================================#
class flash:
    def __init__(self, lat, lon, dt):
        self.lat = lat
        self.lon = lon
        self.dt = dt


# =====================================================================================================================================================================#
class flash_obj:
    def __init__(self, dt, id, flash=None):
        self.modtime = dt
        self.id = id
        if flash:
            self.flashes = [flash]
        else:
            self.flashes = []

    def get_mean_latlon(self):
        nflashes = len(self.flashes)
        latsum = lonsum = 0.0
        for fl in self.flashes:
            latsum += fl.lat
            lonsum += fl.lon
        return latsum / nflashes, lonsum / nflashes

    def get_tai01(self):
        nflashes = len(self.flashes)
        tai = np.zeros(nflashes)
        cter = 0
        for fl in self.flashes:
            tai[cter] = fl.ftai
            cter += 1
        return np.min(tai), np.max(tai)

    def get_nfssigs(self):
        nf1sig = nf2sig = nf3sig = ns1sig = ns2sig = ns3sig = 0
        for fl in self.flashes:
            nf1sig += fl.nf1sig
            nf2sig += fl.nf2sig
            nf3sig += fl.nf3sig
            ns1sig += fl.ns1sig
            ns2sig += fl.ns2sig
            ns3sig += fl.ns3sig
        return nf1sig, nf2sig, nf3sig, ns1sig, ns2sig, ns3sig

    def get_ngse(self):
        nseries = ngroups = nevents = 0
        for fl in self.flashes:
            nseries += fl.nseries
            ngroups += fl.ngroups
            nevents += fl.nevents
        return ngroups, nseries, nevents


# =====================================================================================================================================================================#
def calcDist(lat1, lon1, lat2, lon2):
    R = 6371  # radius of the earth in km
    pi = np.pi
    x = pi / 180 * (lon1 - lon2) * np.cos(0.5 * pi / 180 * (lat1 + lat2))
    y = pi / 180.0 * (lat1 - lat2)
    return R * np.sqrt(x * x + y * y)  # in km


# =====================================================================================================================================================================#
def spherical_dist(pos1, pos2, r=3958.75):
    pos1 = pos1 * 0.017453292519943295  # * np.pi / 180
    pos2 = pos2 * 0.017453292519943295  # *np.pi / 180
    cos_lat1 = np.cos(pos1[..., 0])
    cos_lat2 = np.cos(pos2[..., 0])
    cos_lat_d = np.cos(pos1[..., 0] - pos2[..., 0])
    cos_lon_d = np.cos(pos1[..., 1] - pos2[..., 1])
    return (
        r * np.arccos(cos_lat_d - cos_lat1 * cos_lat2 * (1 - cos_lon_d)) * 1.60934
    )  # conversion from sm to km


# =====================================================================================================================================================================#
def get_fos(flash_objs):

    num_objs = len(flash_objs)

    (
        aid,
        anflashes,
        anf1sig,
        anf2sig,
        anf3sig,
        angroups,
        anseries,
        ans1sig,
        ans2sig,
        ans3sig,
        anevents,
    ) = [np.zeros(num_objs, dtype=np.int32) for x in range(11)]
    alat, alon, atai0, atai1 = [np.empty(num_objs) for x in range(4)]

    cter = 0
    for k in flash_objs:
        aid[cter] = flash_objs[k].id
        alat[cter], alon[cter] = flash_objs[k].get_mean_latlon()
        anflashes[cter] = len(flash_objs[k].flashes)
        atai0[cter], atai1[cter] = flash_objs[k].get_tai01()
        (
            anf1sig[cter],
            anf2sig[cter],
            anf3sig[cter],
            ans1sig[cter],
            ans2sig[cter],
            ans3sig[cter],
        ) = flash_objs[k].get_nfssigs()
        angroups[cter], anseries[cter], anevents[cter] = flash_objs[k].get_ngse()

        cter += 1

    fos = {}
    fos["aid"] = aid
    fos["alat"] = alat
    fos["alon"] = alon
    fos["anflashes"] = anflashes
    fos["angroups"] = angroups
    fos["anseries"] = anseries
    fos["anevents"] = anevents
    fos["atai0"] = atai0
    fos["atai1"] = atai1
    fos["anf1sig"] = anf1sig
    fos["anf2sig"] = anf2sig
    fos["anf3sig"] = anf3sig
    fos["ans1sig"] = ans1sig
    fos["ans2sig"] = ans2sig
    fos["ans3sig"] = ans3sig

    return fos, np.sum(anflashes)


# =====================================================================================================================================================================#
def get_flashstats(flash_objs):

    # get number of total flashes
    nflashes = 0
    for k in flash_objs:
        nflashes += len(flash_objs[k].flashes)

    # create and fill some arrays
    (
        pid,
        fdate,
        ftime,
        nf1sig,
        nf2sig,
        nf3sig,
        ngroups,
        nseries,
        ns1sig,
        ns2sig,
        ns3sig,
        nevents,
    ) = [np.empty(nflashes) for x in range(12)]
    flash_area, flash_energy, lat, lon, duration = [
        np.empty(nflashes) for x in range(5)
    ]
    maxgrpenergy, meangrpenergy, mingrpenergy, maxgrpdis, maxserdur, totserdur = [
        np.empty(nflashes) for x in range(6)
    ]
    maxevtenergy, minevtenergy = [np.empty(nflashes) for x in range(2)]

    cter = 0
    for k in flash_objs:
        for fl in flash_objs[k].flashes:
            pid[cter] = flash_objs[k].id
            fdate[cter] = int(fl.dt.strftime("%Y%m%d"))
            ftime[cter] = int(fl.dt.strftime("%H%M%S"))
            nf1sig[cter] = fl.nf1sig
            nf2sig[cter] = fl.nf2sig
            nf3sig[cter] = fl.nf3sig
            ns1sig[cter] = fl.ns1sig
            ns2sig[cter] = fl.ns2sig
            ns3sig[cter] = fl.ns3sig
            nevents[cter] = fl.nevents
            flash_area[cter] = fl.area
            flash_energy[cter] = fl.energy
            lat[cter] = fl.lat
            lon[cter] = fl.lon
            duration[cter] = fl.duration
            maxgrpenergy[cter] = fl.maxgrpenergy
            meangrpenergy[cter] = fl.meangrpenergy
            mingrpenergy[cter] = fl.mingrpenergy
            maxgrpdis[cter] = fl.maxgrpdis
            maxserdur[cter] = fl.maxserdur
            totserdur[cter] = fl.totserdur
            maxevtenergy[cter] = fl.maxevtenergy
            minevtenergy[cter] = fl.minevtenergy
            cter += 1

    fs = {}
    fs["pid"] = pid
    fs["fdate"] = fdate
    fs["ftime"] = ftime
    fs["lat"] = lat
    fs["lon"] = lon
    fs["duration"] = duration
    fs["nf1sig"] = nf1sig
    fs["nf2sig"] = nf2sig
    fs["nf3sig"] = nf3sig
    fs["ns1sig"] = ns1sig
    fs["ns2sig"] = ns2sig
    fs["ns3sig"] = ns3sig
    fs["nevents"] = nevents
    fs["ngroups"] = ngroups
    fs["area"] = flash_area
    fs["energy"] = flash_energy
    fs["maxgrpenergy"] = maxgrpenergy
    fs["meangrpenergy"] = meangrpenergy
    fs["mingrpenergy"] = mingrpenergy
    fs["maxgrpdis"] = maxgrpdis
    fs["maxserdur"] = maxserdur
    fs["totserdur"] = totserdur
    fs["maxevtenergy"] = maxevtenergy
    fs["minevtenergy"] = minevtenergy

    return fs


# =====================================================================================================================================================================#
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-g",
        "--glmdir",
        help="data directory; default = /arcdata/goes/grb/goes16/",
        type=str,
        default=["/arcdata/goes/grb/goes16/"],
        nargs=1,
    )
    parser.add_argument(
        "-st",
        "--starttime",
        help="starttime (e.g., 20180810-160000; default = now [realtime mode])",
        type=str,
        default=[datetime.utcnow().strftime("%Y%m%d-%H%M%S")],
        nargs=1,
    )
    parser.add_argument(
        "-et",
        "--endtime",
        help="endtime (e.g., 20180811-000000; default = run forever [realtime mode])",
        type=str,
        default=[eternity],
        nargs=1,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="default = ${PWD}",
        type=str,
        default=[os.environ["PWD"] + "/"],
        nargs=1,
    )
    parser.add_argument(
        "-p",
        "--accumulatePeriod",
        help="period to accumulate flashes/areas in minutes; default = 5",
        type=int,
        default=[5],
        nargs=1,
    )
    parser.add_argument(
        "-mnf",
        "--minNumberFlashes",
        help="A flash object must have at least this many recent flashes (see --purgeAge). Default = 5",
        type=int,
        default=[5],
        nargs=1,
    )
    parser.add_argument(
        "-pa",
        "--purgeAge",
        help="Any flashes older than purgeAge seconds will be removed from flash object. Default = 300 s",
        type=int,
        default=[300],
        nargs=1,
    )
    parser.add_argument(
        "-eps",
        "--epsilon",
        help="A parameter for DBSCAN. Default = 0.02. For larger objects, try 0.5. For smaller objects, try 0.015.",
        type=float,
        default=[0.02],
        nargs=1,
    )
    parser.add_argument(
        "-b",
        "--buffer",
        help="Add a buffer around objects when computing overlap (for tracking). This is in lat/lon degrees. Default = 0.1",
        type=float,
        default=[0.1],
        nargs=1,
    )
    parser.add_argument(
        "-l",
        "--logfile",
        help="Logfile. Default = None.",
        default=["None"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-j", "--jsons", help="Convert to jsons. Default = False.", action="store_true"
    )

    args = parser.parse_args()

    logfile = None if (args.logfile[0] == "None") else args.logfile[0]

    status = glmobj_dbscan(
        args.glmdir[0],
        args.starttime[0],
        args.endtime[0],
        outdir=args.outdir[0],
        period=args.accumulatePeriod[0],
        epsilon=args.epsilon[0],
        minNumberFlashes=args.minNumberFlashes[0],
        purgeAge=args.purgeAge[0],
        jsons=args.jsons,
        logfile=logfile,
        buffer=args.buffer[0],
    )


# =====================================================================================================================================================================#
def glmobj_dbscan(
    glmdir,
    starttime=datetime.utcnow().strftime("%Y%m%d-%H%M%S"),
    endtime=eternity,
    outdir=os.environ["PWD"] + "/",
    period=5,
    minNumberFlashes=5,
    purgeAge=300,
    epsilon=0.02,
    buffer=0.1,
    jsons=False,
    **kwargs
):

    FUNCTION_NAME = "GLMOBJ_DBSCAN"
    if "logfile" in kwargs:
        logfile = kwargs["logfile"]
        if logfile:
            utils.mkdir_p(os.path.dirname(logfile))
    else:
        logfile = None
    logDict = logging_def(logfile=logfile)
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    logging.info(FUNCTION_NAME + ": Process started")

    epoch = datetime.utcfromtimestamp(0)  # 1970/01/01 00:00

    try:
        dt = datetime.strptime(starttime, "%Y%m%d-%H%M%S")
        edt = datetime.strptime(endtime, "%Y%m%d-%H%M%S")
    except ValueError as err:
        logging.error(FUNCTION_NAME + ": " + traceback.format_exc())
        return False

    # round dt to nearest 20sec
    ses = int((dt - epoch).total_seconds())
    rounded_ses = np.round(20 * np.round(float(ses) / 20))
    dt = datetime.utcfromtimestamp(rounded_ses)

    realtime = True if (edt.strftime("%Y%m%d-%H%M%S") == eternity) else False

    maxid = 9999

    wait_time = nobjs = total_nflashes = 0
    all_flashes = []

    while dt < edt:
        # dt is iterated below, adding 20s each time
        successful_iteration = False

        # find ceil of nearest 'period' min
        es = int((dt - epoch).total_seconds())
        rounded_es = np.round((60 * period) * np.ceil(float(es) / (period * 60.0)))
        write_dt = datetime.utcfromtimestamp(rounded_es)

        # open files, aggregate stats
        ncfile = os.path.join(
            glmdir,
            dt.strftime("%Y/%Y_%m_%d_%j/glm/L2/LCFA/OR_GLM-L2-LCFA_G16_s%Y%j%H%M%S*"),
        )
        ncfiles = glob.glob(ncfile)
        if len(ncfiles) == 1:
            logging.info(
                FUNCTION_NAME
                + ": Processing Datetime "
                + dt.strftime("%Y-%m-%d %H:%M:%S")
            )
            ncfile = ncfiles[0]
            for attempt in range(5):
                try:
                    shutil.copy(ncfile, os.environ["PWD"])
                    basencfile = os.path.basename(ncfile)
                    glm = Dataset(basencfile, "r")
                    fid = np.where(glm.variables["flash_quality_flag"][:] == 0)[0]
                    cnf = len(fid)
                except (OSError, IOError, ValueError) as err:
                    logging.error(FUNCTION_NAME + ": " + traceback.format_exc())
                    time.sleep(5)
                    wait_time += 5
                    cnf = 0

            if cnf:
                logging.info(FUNCTION_NAME + ": Iterating through flashes...")
                fdate = np.zeros(cnf, dtype=np.int32) + int(dt.strftime("%Y%m%d"))
                ftime = np.zeros(cnf, dtype=np.int32) + int(dt.strftime("%H%M%S"))
                product_time = glm.variables["product_time"][:]
                flash_time_offset_of_first_event = glm.variables[
                    "flash_time_offset_of_first_event"
                ][fid]
                flash_time_offset_of_last_event = glm.variables[
                    "flash_time_offset_of_last_event"
                ][fid]
                flash_area = glm.variables["flash_area"][fid]
                flash_energy = glm.variables["flash_energy"][fid]
                group_energy = glm.variables["group_energy"][:]
                event_energy = glm.variables["event_energy"][:]
                lat = glm.variables["flash_lat"][fid]
                lon = glm.variables["flash_lon"][fid]
                flash_id = glm.variables["flash_id"][:]
                glat = glm.variables["group_lat"][:]
                glon = glm.variables["group_lon"][:]
                group_id = glm.variables["group_id"][:]
                group_time_offset = glm.variables["group_time_offset"][:]
                group_parent_flash_id = glm.variables["group_parent_flash_id"][:]
                event_parent_group_id = glm.variables["event_parent_group_id"][:]
                glm.close()

                os.remove(basencfile)

                ftai = product_time + 2 * flash_time_offset_of_first_event / 1000.0
                duration = (
                    (flash_time_offset_of_last_event - flash_time_offset_of_first_event)
                    * 2
                    / 1000.0
                )

                # for each flash
                for j in range(cnf):
                    gid = np.where(group_parent_flash_id == flash_id[fid[j]])[0]
                    cng = len(gid)
                    if cng < 5:
                        continue
                    eid = []
                    # for each group in the flash, get the events
                    for k in range(cng):
                        ide = np.where(event_parent_group_id == group_id[gid[k]])[0]
                        if len(ide) == 0:
                            continue  # can't find parent group?
                        eid = np.append(eid, ide)
                    if len(eid) < 5:
                        continue
                    cne = len(eid)
                    if cne:
                        eid = eid.astype(int)

                    new_flash = flash(lat[j], lon[j], dt)

                    new_flash.energy = flash_energy[j]
                    new_flash.area = flash_area[j]
                    new_flash.duration = duration[j]
                    new_flash.ftai = ftai[j]
                    new_flash.ngroups = cng
                    new_flash.maxgrpenergy = max(
                        group_energy[gid]
                    )  # *1.52597e-15+3.e-15)
                    new_flash.meangrpenergy = np.mean(
                        group_energy[gid]
                    )  # *1.52597e-15+3.e-15)
                    new_flash.mingrpenergy = min(
                        group_energy[gid]
                    )  # *1.52597e-15+3.e-15)
                    new_flash.nevents = cne
                    new_flash.maxevtenergy = max(
                        event_energy[eid]
                    )  # *1.52597e-15+3.e-15)
                    new_flash.minevtenergy = min(
                        event_energy[eid]
                    )  # *1.52597e-15+3.e-15)
                    ind = np.where(
                        group_energy[gid]
                        > np.mean(group_energy[gid]) + 1.0 * np.std(group_energy[gid])
                    )[0]
                    new_flash.nf1sig = len(ind)
                    ind = np.where(
                        group_energy[gid]
                        > np.mean(group_energy[gid]) + 2.0 * np.std(group_energy[gid])
                    )[0]
                    new_flash.nf2sig = len(ind)
                    ind = np.where(
                        group_energy[gid]
                        > np.mean(group_energy[gid]) + 3.0 * np.std(group_energy[gid])
                    )[0]
                    new_flash.nf3sig = len(ind)

                    dis1 = calcDist(lat[j], lon[j], glat[gid], glon[gid])
                    id1 = np.where(dis1 == max(dis1))[0][0]
                    dis2 = calcDist(
                        glat[gid[id1]], glon[gid[id1]], glat[gid], glon[gid]
                    )
                    id2 = np.where(dis2 == max(dis2))[0][0]

                    new_flash.maxgrpdis = calcDist(
                        glat[gid[id1]], glon[gid[id1]], glat[gid[id2]], glon[gid[id2]]
                    )

                    # calcs for maxevtdis?

                    scount = 0
                    new_flash.nseries = 0
                    new_flash.totserdur = 0
                    new_flash.maxserdur = 0
                    new_flash.ns1sig = 0
                    new_flash.ns2sig = 0
                    new_flash.ns3sig = 0
                    # for g in range(cng):
                    #   news = 0
                    #   if(g == 0): news = 1
                    #   if(g > 0):
                    #     if(abs(group_time_offset[gid[g]]*2-group_time_offset[gid[g-1]]*2) > 2.1*1): news = 1

                    #   if (news == 1):
                    # This group should be in a new series....
                    # Are we working on a series?
                    #     if(scount > 0):

                    #       new_flash.nseries += 1
                    #       dtime = (max(group_time_offset[sgid]*2)-min(group_time_offset[sgid]*2))/1000.
                    #       new_flash.totserdur += dtime
                    #       if (dtime > new_flash.maxserdur): new_flash.maxserdur = dtime
                    #       ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+1.*np.std(group_energy[gid]))[0]
                    #       if (len(ind)): new_flash.ns1sig += 1
                    #       ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+2.*np.std(group_energy[gid]))[0]
                    #       if (len(ind)): new_flash.ns2sig += 1
                    #       ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+3.*np.std(group_energy[gid]))[0]
                    #       if (len(ind)): new_flash.ns3sig += 1

                    #     scount = 1
                    #     sgid = [gid[g]]

                    #   if (news == 0):
                    #     sgid.append(gid[g])
                    #     scount += 1

                    #  new_flash.nseries += 1
                    #  dtime = (max(group_time_offset[sgid]*2)-min(group_time_offset[sgid]*2))/1000.
                    #  new_flash.totserdur += dtime
                    #  if (dtime > new_flash.maxserdur): new_flash.maxserdur = dtime
                    #  ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+1.*np.std(group_energy[gid]))[0]
                    #  if (len(ind)): new_flash.ns1sig += 1
                    #  ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+2.*np.std(group_energy[gid]))[0]
                    #  if (len(ind)): new_flash.ns2sig += 1
                    #  ind = np.where(group_energy[sgid] > np.mean(group_energy[gid])+3.*np.std(group_energy[gid]))[0]
                    #  if (len(ind)): new_flash.ns3sig += 1

                    all_flashes.append(new_flash)

                # ======================================#
                # ==done iterating through new flashes==#
                # ======================================#

                # Purge old flashes
                logging.info("purging old flashes...")
                for fl in all_flashes:
                    if (dt - fl.dt).total_seconds() > purgeAge:
                        all_flashes.remove(fl)
                logging.info("done purging")
            else:
                logging.warning(FUNCTION_NAME + ": No good flashes present.")
                # just close the netcdf
                try:
                    glm.close()
                except (OSError, IOError, ValueError, RuntimeError) as err:
                    logging.warning(FUNCTION_NAME + ": " + traceback.format_exc())
                    pass

            # reset as long as we successfully found and read the file for this dt
            if cnf:
                successful_iteration = True
                wait_time = 0

        else:
            # file was not found. If realtime, go ahead and take a nap.
            if realtime:
                # pass
                logging.info(
                    FUNCTION_NAME
                    + ": "
                    + ncfile
                    + " is not available (wait_time = "
                    + str(wait_time)
                    + " sec)"
                )
            else:
                logging.warning(FUNCTION_NAME + ": " + ncfile + " was not found.")

        if dt == write_dt and successful_iteration:

            # Now use DBSCAN to make objects
            # min_samples = minimum number of flashes in an object
            logging.info(FUNCTION_NAME + ": Running DBSCAN")
            nflashes = len(all_flashes)
            latlon = np.zeros((nflashes, 2))
            for ff in range(nflashes):
                latlon[ff] = all_flashes[ff].lat, all_flashes[ff].lon

            X = StandardScaler().fit_transform(latlon)
            db = DBSCAN(eps=epsilon, min_samples=minNumberFlashes).fit(X)
            ids = np.unique(db.labels_)
            ids = ids[ids >= 0]

            # assign each flash the parent objectID, and assign the flash to that flash_obj
            # these are for temporary ids
            # first make list of flashless flash_objs
            flash_objs = {}
            for id in ids:
                flash_objs[id] = flash_obj(dt, id)
            for ff in range(nflashes):
                all_flashes[ff].pid = db.labels_[ff]
                if (
                    db.labels_[ff] >= 0
                ):  # labels of -1 mean they were not assigned to an object
                    flash_objs[db.labels_[ff]].flashes.append(all_flashes[ff])
            logging.info(FUNCTION_NAME + ": Getting stats for the tmp flash objects...")
            tmp_flash_obj_stats, total_nflashes = get_fos(
                flash_objs
            )  # get "flash object stats"
            tmp_flash_stats = get_flashstats(flash_objs)

            # ============#
            # ==Tracking==#
            # ============#
            logging.info(FUNCTION_NAME + ": Tracking objects...")

            # use object overlap to do tracking
            try:
                final_flash_obj_stats
            except NameError as err:
                logging.info(FUNCTION_NAME + ": First output time...")
                # the first time
                # instantiate final_flash_obj_stats with tmp_flash_objs_stats; final_flash_stats with tmp_flash_stats; final_flash_objs with flash_objs
                final_flash_obj_stats = copy.deepcopy(tmp_flash_obj_stats)
                final_flash_stats = copy.deepcopy(tmp_flash_stats)
                final_flash_objs = copy.deepcopy(flash_objs)
                # start idcter
                idcter = np.max(db.labels_)
            else:
                # not the first time
                # get centroids of objs from previous times
                prev_mlatlons = np.array(
                    [
                        [hh, gg]
                        for hh, gg in zip(
                            prev_flash_obj_stats["alat"], prev_flash_obj_stats["alon"]
                        )
                    ]
                )

                # for each current (tmp) obj, find nearby current objs
                # use the id of the max overlap, or create a new id
                final_flash_objs = {}

                currID, IDused, overlapID, overlapArea = [
                    np.zeros(len(flash_objs)) - 999 for x in range(4)
                ]
                fo_cter = 0

                for (
                    id
                ) in (
                    flash_objs
                ):  # these are flash objects at current time, with DBSCAN ids (i.e., 0 to nobjs)
                    mlat, mlon = flash_objs[id].get_mean_latlon()
                    distInKm = spherical_dist(np.array([mlat, mlon]), prev_mlatlons)
                    cand_ind = np.where(distInKm < 250)

                    if np.size(cand_ind):

                        # since we have some nearby previous objs, make convex_hull for this curr obj
                        cand_flash_coords = []
                        for ff in flash_objs[id].flashes:
                            cand_flash_coords.append((ff.lon, ff.lat))
                        curr_ch = MultiPoint(cand_flash_coords).convex_hull.buffer(
                            buffer, cap_style=1, join_style=1
                        )

                        # find previous obj with max overlap
                        maxarea = 0
                        for cand in cand_ind[0]:
                            candid = int(final_flash_obj_stats["aid"][cand])
                            cand_flash_coords = []
                            flash_ind = np.where(final_flash_stats["pid"] == candid)
                            for ff in flash_ind[0]:
                                cand_flash_coords.append(
                                    (
                                        final_flash_stats["lon"][ff],
                                        final_flash_stats["lat"][ff],
                                    )
                                )
                            cand_ch = MultiPoint(cand_flash_coords).convex_hull.buffer(
                                buffer, cap_style=1, join_style=1
                            )

                            tmparea = curr_ch.intersection(cand_ch).area
                            if tmparea > maxarea:
                                maxarea = tmparea
                                previd = candid

                        currID[fo_cter] = id
                        overlapID[fo_cter] = previd
                        overlapArea[fo_cter] = maxarea
                        fo_cter += 1

                    else:
                        # no candidates within distInKm. Set to new ID.
                        IDused[fo_cter] = 1
                        fo_cter += 1  # this is just so we don't needlessly iterate over flash_objs that don't have any matches
                        # no previous objs in distance window; assign a new id
                        idcter = (idcter + 1) % (maxid + 1)
                        final_flash_objs[idcter] = flash_objs[id]
                        final_flash_objs[idcter].id = idcter

                for o in range(len(overlapID)):
                    if IDused[o] < 0:
                        good_inds = np.where(overlapID == overlapID[o])
                        if np.size(good_inds) > 1:
                            max_overlap = 0
                            for qq in range(np.size(good_inds)):
                                IDused[good_inds[0][qq]] = 1
                                if overlapArea[good_inds[0][qq]] > max_overlap:
                                    # keep overwritting until we have the best overlap.
                                    max_overlap = overlapArea[good_inds[0][qq]]
                                    final_flash_objs[overlapID[o]] = flash_objs[
                                        currID[good_inds[0][qq]]
                                    ]
                                    final_flash_objs[overlapID[o]].id = overlapID[o]
                                    best_ind = qq
                            # now go through the good_inds again, and assign a new id to the flash obj(s) that didn't have the max overlap (i.e., != best_ind)
                            for qq in range(np.size(good_inds)):
                                if qq != best_ind:
                                    idcter = (idcter + 1) % (maxid + 1)
                                    final_flash_objs[idcter] = flash_objs[
                                        currID[good_inds[0][qq]]
                                    ]
                                    final_flash_objs[idcter].id = idcter

                        else:
                            # only 1 match. Set to the a current flash obj and set id to the overlap id
                            IDused[o] = 1
                            final_flash_objs[overlapID[o]] = flash_objs[currID[o]]
                            final_flash_objs[overlapID[o]].id = overlapID[o]

            logging.info(FUNCTION_NAME + ": done tracking objects")

            # Now get summary stats for flashes and objects
            logging.info(
                FUNCTION_NAME + ": Getting stats for the final flash objects..."
            )
            final_flash_obj_stats, total_nflashes = get_fos(
                final_flash_objs
            )  # get "flash object stats"
            logging.info(FUNCTION_NAME + ": Getting stats for the flashes...")
            final_flash_stats = get_flashstats(final_flash_objs)

            prev_flash_obj_stats = copy.deepcopy(final_flash_obj_stats)

            nobjs = len(final_flash_objs)

            logging.info(
                FUNCTION_NAME
                + ": Number of flashes = "
                + str(total_nflashes)
                + "; Number of areas = "
                + str(nobjs)
            )

            if total_nflashes > 0 and nobjs > 0:
                logging.info(
                    FUNCTION_NAME
                    + ": Writing data for last "
                    + str(period)
                    + " min ending at "
                    + dt.strftime("%Y-%m-%d %H:%M:%S")
                )
                timestamp = dt.strftime("%Y%m%d-%H%M%S")
                outfile = os.path.join(
                    outdir,
                    "goes_east",
                    timestamp[0:8],
                    "glmobjs",
                    timestamp + ".netcdf",
                )

                dims = OrderedDict()
                dims["nflashes"] = total_nflashes
                dims["nareas"] = nobjs
                dataset = OrderedDict()
                for varname in final_flash_stats:
                    dataset[varname] = {
                        "data": final_flash_stats[varname],
                        "dims": "nflashes",
                        "atts": {},
                    }
                for varname in final_flash_obj_stats:
                    dataset[varname] = {
                        "data": final_flash_obj_stats[varname],
                        "dims": "nareas",
                        "atts": {},
                    }

                atts = OrderedDict()
                atts["timestamp"] = timestamp + " UTC"
                atts["period"] = str(period) + " min"
                atts["AreaIdentificationMethod"] = "DBSCAN"
                atts["purgeAge"] = str(purgeAge) + " sec"
                atts["minNumberFlashesPerArea"] = str(minNumberFlashes)

                # write netcdf
                status = utils.write_netcdf(
                    outfile, dataset, dims, atts=atts, gzip=False
                )

                # write json
                if jsons:
                    logging.info(FUNCTION_NAME + ": Making json for RealEarth")
                    # data = {}
                    # for key in flash_stats:
                    #   data[key] = flash_stats[key]
                    # for key in flash_obj_stats:
                    #   data[key] = flash_obj_stats[key]
                    # data['timestamp'] = timestamp
                    # status = glm_objs_json(data=data,outdir=os.path.join(outdir,'goes_east',timestamp[0:8],'jsons'))

                    status = Popen(
                        [
                            "python",
                            "glm_objs_json.py",
                            "-f",
                            outfile,
                            "-o",
                            os.path.join(outdir, "goes_east", timestamp[0:8], "jsons"),
                            "-u",
                        ]
                    )

                    # upload to RealEarth
                    # if(upload):
                    # re_upload_status = os.system('./re_upload ' + os.path.join(outdir,'goes_east',timestamp[0:8],'jsons','GLMOBJ_'+timestamp[0:8]+'_'+timestamp[9:15]+'.force.json'))
                    # if(re_upload_status != 0): logging.error(FUNCTION_NAME + ': RealEarth upload failed')

                    #  status = Popen(['./re_upload',os.path.join(outdir,'goes_east',timestamp[0:8],'jsons','GLMOBJ_'+timestamp[0:8]+'_'+timestamp[9:15]+'.force.json')])

        # increment
        if realtime:
            if successful_iteration:
                dt += timedelta(seconds=20)
            elif wait_time >= 120:  # if we've waited for 2 min, reset
                # grab latest file in glmdir
                dtnow = datetime.utcnow()
                today_dir = os.path.join(
                    glmdir,
                    dtnow.strftime("%Y/%Y_%m_%d_%j/glm/L2/LCFA/OR_GLM-L2-LCFA_G16_*"),
                )
                latest_file = None
                try:
                    latest_file = np.sort(glob.glob(today_dir))[-1]
                except IndexError as err:
                    logging.error(FUNCTION_NAME + ": " + traceback.format_exc())
                    yesterday = dtnow - timedelta(days=1)
                    yesterday_dir = os.path.join(
                        glmdir,
                        yesterday.strftime(
                            "%Y/%Y_%m_%d_%j/glm/L2/LCFA/OR_GLM-L2-LCFA_G16_*"
                        ),
                    )
                    try:
                        latest_file = np.sort(glob.glob(yesterday_dir))[-1]
                    except IndexError as err:
                        logging.error(FUNCTION_NAME + ": " + traceback.format_exc())
                        logging.warning(
                            FUNCTION_NAME + ": Something may be wrong with GLMDIR."
                        )
                        time.sleep(5)

                try:
                    basefile = os.path.basename(latest_file)
                    yyyyjjjhhmmss = basefile.split("_s")[1][0:13]
                    dt = datetime.strptime(yyyyjjjhhmmss, "%Y%j%H%M%S")
                    logging.warning(
                        FUNCTION_NAME
                        + ": Waited for "
                        + str(wait_time)
                        + " sec. Now using dt from "
                        + latest_file
                    )
                except (NameError, TypeError, ValueError, IndexError) as err:
                    logging.error(FUNCTION_NAME + ": " + traceback.format_exc())

                # reset wait_time
                wait_time = 0

            else:
                time.sleep(3)
                wait_time += 3
        else:
            # in offline mode, increment whether or not it was a successful iteration. There may just be missing files.
            dt += timedelta(seconds=20)

    logging.info(FUNCTION_NAME + ": Process ended")


if __name__ == "__main__":
    main()
