from __future__ import absolute_import, division, print_function, unicode_literals

import traceback
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import os, sys

pltg_data = os.environ["PLTG_DATA"] + "/"
pltg = os.environ["PLTG"] + "/"
sys.path.insert(0, os.path.join(pltg, "src", "static/"))
from scipy.ndimage import zoom

# os.environ['CUDA_DEVICE_ORDER']="PCI_BUS_ID"
# os.environ["CUDA_VISIBLE_DEVICES"]=""
from netCDF4 import Dataset
import numpy as np
import argparse
import pandas as pd

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model, save_model, load_model

from train_eval_cnn import normalize_multivariate_data as nmd
from lightningcast import keras_metrics
import matplotlib.pyplot as plt
import time
from subprocess import PIPE, call, Popen
from pyproj import Proj
from ABIcmaps import irw_ctc
from lightningcast import utils
import numpy.ma as ma
from datetime import datetime, timedelta, timezone
import re
from lightningcast.solar_angles import solar_angles
from lightningcast.sat_zen_angle import sat_zen_angle
import pickle
import glob
from matplotlib.lines import Line2D
import shutil
from collections import OrderedDict
from lightningcast.collect_data import read_netcdf
from DataGenerator import DataGeneratorPredict
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
from lightningcast.earth_to_fgf import earth_to_fgf

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

LIST_OF_METRIC_FUNCTIONS = [
    keras_metrics.accuracy,
    keras_metrics.binary_accuracy,
    keras_metrics.binary_csi,
    keras_metrics.binary_frequency_bias,
    keras_metrics.binary_pod,
    keras_metrics.binary_pofd,
    keras_metrics.binary_peirce_score,
    keras_metrics.binary_success_ratio,
    keras_metrics.binary_focn,
]
METRIC_FUNCTION_DICT = {
    "accuracy": keras_metrics.accuracy,
    "binary_accuracy": keras_metrics.binary_accuracy,
    "binary_csi": keras_metrics.binary_csi,
    "binary_frequency_bias": keras_metrics.binary_frequency_bias,
    "binary_pod": keras_metrics.binary_pod,
    "binary_pofd": keras_metrics.binary_pofd,
    "binary_peirce_score": keras_metrics.binary_peirce_score,
    "binary_success_ratio": keras_metrics.binary_success_ratio,
    "binary_focn": keras_metrics.binary_focn,
}
#################################################################################################################################
def get_rgb_colors(colors):
    # colors list of hex strings
    rgb_colors = []
    for c in colors:
        h = c.lstrip("#")
        hlist = [str(int(h[i : i + 2], 16)) for i in (0, 2, 4)]
        rgb_colors.append((" ").join(hlist))

    return rgb_colors


#################################################################################################################################
from shapely import geometry
from geojson import Polygon, Feature, FeatureCollection
import geojson
from geojson_rewind import rewind
import json


def write_json(
    cs,
    outfile,
    re_upload=None,
    values=[0.25, 0.5, 0.75],
    colors=["0 0 255", "0 255 255", "255 0 255"],
    empty=False,
):
    fc_list = []
    cter = 0

    if empty:
        fc_list = []
    else:
        for col in cs.collections:
            # Loop through all polygons that have the same intensity level
            for contour_path in col.get_paths():
                # Create the polygon for this intensity level
                # The first polygon in the path is the main one, the following ones are "holes"

                for ncp, cp in enumerate(contour_path.to_polygons()):
                    # print(cp)
                    try:
                        x = cp[:, 0]
                        y = cp[:, 1]
                    except TypeError:
                        # print('TypeError')
                        x = [a[0] for a in cp]
                        y = [a[1] for a in cp]
                    new_shape = geometry.Polygon(
                        [(round(i[0], 3), round(i[1], 3)) for i in zip(x, y)]
                    )
                    if ncp == 0:
                        poly = new_shape
                    else:
                        # Remove the holes if there are any
                        poly = poly.difference(new_shape)
                        # Can also be left out if you want to include all rings

                    # do something with polygon
                    poly = Feature(geometry=poly)
                    poly["properties"]["OPACITY"] = "0"
                    poly["properties"]["BOPACITY"] = "100"
                    poly["properties"]["COLOR"] = colors[cter]
                    poly["properties"]["BCOLOR"] = colors[cter]
                    poly["properties"]["WIDTH"] = "1"
                    poly["properties"]["INFO"] = (
                        "ICP: " + str(int(values[cter] * 100)) + "%"
                    )
                    if poly.is_valid:
                        fc_list.append(poly)
                    else:
                        logging.error("poly not valid:")
                        print(poly)

            cter += 1

    fc = FeatureCollection(fc_list)
    fc_str = rewind(fc)  # makes json 'right-hand rule' compliant
    fc_str = json.dumps(fc_str, indent=2, sort_keys=True)  # pretty printing
    utils.mkdir_p(os.path.dirname(outfile))
    of = open(outfile, "w")
    of.write(fc_str)
    of.close()

    if (
        re_upload
    ):  # re_upload is either None or a string to full path of the re_upload script
        s = call([re_upload, outfile])


#################################################################################################################################
def find_latest(listened_file, sector):
    if not (os.path.isfile(listened_file)):
        logging.critical(listened_file + " is not a file or does not exist")
        sys.exit(1)
    f = open(listened_file, "r")
    lines = f.read().split("\n")
    logging.info("Checking lines for " + sector)
    latest = None
    nlines = len(lines)
    for i in range(nlines - 1, 0, -1):
        if sector in lines[i]:
            latest = lines[i]
            break

    if latest:
        parts = latest.split("#")
        machine = parts[1]
        partition = "/" + machine.split(".")[0] + "_data"
        filestr = parts[-1]
        basefile = os.path.basename(filestr)
        fileparts = filestr.split("_")
        timestamp = fileparts[-1][1:14]
        dt = datetime.strptime(timestamp, "%Y%j%H%M%S")
        remoteDir = (os.path.dirname(filestr)).replace(
            "/data", partition
        )  # should be '/satbuf1_data', or '/satbuf2_data'
    else:
        basefile = remoteDir = dt = "None"

    return basefile, remoteDir, dt


#################################################################################################################################
def find_corners(input_tuples, input_list):
    input_corners = (
        []
    )  # a list of lists of tuples for each NW corner point to feed to the model. Should be len(input_list)
    # This loop finds the correct "northwest" corner points for each model input for each large patch (in data space).
    for inp in range(len(input_tuples)):

        patch_ny = input_tuples[inp][0]
        patch_nx = input_tuples[inp][1]
        input_ny, input_nx = input_list[inp][0].shape

        #  ny_corners = input_ny // patch_ny
        #  y_corners = [patch_ny * _ for _ in range(ny_corners+1)]
        y_corners = list(np.arange(0, input_ny, patch_ny // 2))
        # interp y_corners
        #  y_corners = list(np.interp(np.arange(ny_corners*2+1),np.arange(0,ny_corners*2+1,2),y_corners).astype(int))
        # remainder

        y_remainder = input_ny % patch_ny
        if y_remainder > 0:
            last_corner = (
                input_ny - patch_ny
            )  # position last corner so it lands on input image edge
            if last_corner < y_corners[-2]:
                _ = y_corners.pop(
                    -1
                )  # remove last element, as we need to overwrite second to last
            y_corners[-1] = last_corner

        #   nx_corners = input_nx // patch_nx
        #   x_corners = [patch_nx * _ for _ in range(nx_corners+1)]
        x_corners = list(np.arange(0, input_nx, patch_nx // 2))
        # interp y_corners
        #   x_corners = list(np.interp(np.arange(nx_corners*2+1),np.arange(0,nx_corners*2+1,2),x_corners).astype(int))
        # remainder
        x_remainder = input_nx % patch_nx
        if x_remainder > 0:
            last_corner = (
                input_nx - patch_nx
            )  # position last corner so it lands on input image edge
            if last_corner < x_corners[-2]:
                _ = x_corners.pop(
                    -1
                )  # remove last element, as we need to overwrite second to last
            x_corners[-1] = last_corner

        corners = []
        for yc in y_corners:
            corners += [(i, j) for i, j in zip([yc] * len(x_corners), x_corners)]
        input_corners.append(corners)

    return input_corners, y_remainder, x_remainder


#################################################################################################################################
def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx, array[idx]


#################################################################################################################################
def get_reps(reports_file):
    lines = [line.rstrip("\n") for line in open(reports_file)]
    reps = []
    init_dt = datetime.strptime("20" + reports_file[0:6], "%Y%m%d")
    for ll in lines:
        if "Time," in ll:
            continue
        parts = ll.split(",")  # parts[0] = HHMM, parts[-2] = lon, parts[-3] = lat
        hhmm = parts[0]
        td = timedelta(hours=int(hhmm[0:2]), minutes=int(hhmm[2:4]))
        if int(hhmm) >= 1200:
            dt = init_dt + td
        else:
            dt = init_dt + td + timedelta(hours=24)
        reps.append((dt, float(parts[-2]), float(parts[-3])))
    return reps


#################################################################################################################################
def get_abi_data(
    dt,
    ch,
    remotedir,
    sector,
    startY=None,
    endY=None,
    startX=None,
    endX=None,
    metadata={},
    return_proj_info=False,
):
    # returns the 2D reflectance or BT array

    try:
        remotefile = glob.glob(
            dt.strftime(
                os.path.join(
                    remotedir,
                    "%Y",
                    "%Y_%m_%d_%j",
                    "abi",
                    "L1b",
                    sector,
                    "*C" + ch + "_*_s%Y%j%H%M*",
                )
            )
        )[0]
    except IndexError:
        raise
    else:
        # copy file to $PWD
        bf = os.path.basename(remotefile)
        localfile = os.path.join(os.environ["PWD"], bf)
        shutil.copy(remotefile, localfile)
        nc = Dataset(localfile, "r")

        # get the data
        radiance = nc.variables["Rad"][:]

        metadata["goes_position"] = nc.orbital_slot
        metadata["scene_id"] = nc.scene_id
        # tce = datetime.strptime(nc.time_coverage_end[0:16],'%Y-%m-%dT%H:%M')

        if ch == "02":
            # convert vis radiance to reflectance
            result = radiance * nc.variables["kappa0"][:]
            # crop, if necessary
            if startY:
                result = result[startY * 4 : endY * 4, startX * 4 : endX * 4]
        elif ch in ["01", "03", "05"]:
            # convert vis radiance to reflectance
            result = radiance * nc.variables["kappa0"][:]
            # crop, if necessary
            if startY:
                result = result[startY * 2 : endY * 2, startX * 2 : endX * 2]
        else:  # IR channels
            # convert IR radiance to BT
            result = (
                nc.variables["planck_fk2"][:]
                / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                - nc.variables["planck_bc1"][:]
            ) / nc.variables["planck_bc2"][:]
            # crop, if necessary
            if startY:
                result = result[startY:endY, startX:endX]

        if return_proj_info:
            satheight = nc.variables["goes_imager_projection"].getncattr(
                "perspective_point_height"
            )
            x = nc.variables["x"][:] * satheight
            y = nc.variables["y"][:] * satheight

        try:
            nc.close()
        except RuntimeError:
            pass

        try:
            os.remove(localfile)
        except (NameError, IOError, OSError):
            pass

    if return_proj_info:
        return x, y, result
    else:
        return result


#################################################################################################################################
def get_meso_proj_info(dt, remote_dir, sector):
    for ch in ["02", "13"]:
        try:
            remotefile = glob.glob(
                dt.strftime(
                    os.path.join(
                        remote_dir,
                        "%Y",
                        "%Y_%m_%d_%j",
                        "abi",
                        "L1b",
                        sector,
                        "*C" + ch + "_*_s%Y%j%H%M*",
                    )
                )
            )[0]
        except IndexError:
            raise
        else:
            # copy file to $PWD
            bf = os.path.basename(remotefile)
            localfile = os.path.join(os.environ["PWD"], bf)
            shutil.copy(remotefile, localfile)
            nc = Dataset(localfile, "r")

            satheight = nc.variables["goes_imager_projection"].getncattr(
                "perspective_point_height"
            )

            # get the data
            if ch == "02":
                visx = nc.variables["x"][:] * satheight
                visy = nc.variables["y"][:] * satheight
            else:
                irx = nc.variables["x"][:] * satheight
                iry = nc.variables["y"][:] * satheight

            try:
                nc.close()
            except RuntimeError:
                pass

            try:
                os.remove(localfile)
            except (NameError, IOError, OSError):
                pass

    return visx, visy, irx, iry


#################################################################################################################################
def predict_ltg(
    filelist,
    model,
    model_config_file,
    outdir=os.environ["PWD"] + "/OUTPUT/",
    glmdir=None,
    bounding_box=[-999],
    remote_dir=None,
    sandimg=False,
    alpha=0.25,
    values=[0.25, 0.5, 0.9],
    sector="RadC",
    savegrid=False,
    realtime=False,
    slot="GOES_East",
    re_upload=None,
    make_json=False,
):

    utils.mkdir_p(outdir)
    t00 = time.time()

    state_borders = cfeature.NaturalEarthFeature(
        category="cultural",
        name="admin_1_states_provinces_lakes",
        scale="50m",
        facecolor="none",
    )

    conv_model = load_model(model, custom_objects=METRIC_FUNCTION_DICT, compile=False)

    mc = pickle.load(open(model_config_file, "rb"))
    svals = mc["scaling_values"]
    channels = mc["channels"]
    ninputs = len(mc["input_tuples"])
    scalar_vars = mc["scalar_vars"]
    inputs = mc["inputs"]

    # with respect to ABI CONUS IR grid #bounding_box=[SWlon, NElon, SWlat, NElat]
    # meso or CONUS? GE or GW?
    # get filenames
    if realtime:
        nlines = 1
        latest_filestr = filestr = "foo"
    else:
        slot = "GOES_East" if ("goes16" in remote_dir) else "GOES_West"
    currhrrrfile = "foo"
    glmfile = None

    # get file times
    with open(filelist) as fp:
        lines = [l.rstrip("\n") for l in fp.readlines()]
    fp.close()
    nlines = len(lines)

    # static 2km lats, lons
    conus2km = Dataset(os.path.join(pltg, "src", "static", slot + ".nc"), "r")
    satheight = conus2km.variables["goes_imager_projection"].getncattr(
        "perspective_point_height"
    )
    lon0 = conus2km.variables["goes_imager_projection"].getncattr(
        "longitude_of_projection_origin"
    )
    sweep = conus2km.variables["goes_imager_projection"].getncattr("sweep_angle_axis")
    # projections
    p = Proj(proj="geos", h=satheight, lon_0=lon0, sweep=sweep)
    geoproj = ccrs.Geostationary(
        central_longitude=lon0, sweep_axis=sweep, satellite_height=satheight
    )

    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    irx = conus2km.variables["x"][:] * satheight
    iry = conus2km.variables["y"][:] * satheight
    irxx, iryy = np.meshgrid(irx, iry)
    irlons, irlats = p(irxx, iryy, inverse=True)
    conus2km.close()

    # static 05km lats, lons
    # conus05km = Dataset(os.path.join(pltg,'src','static',slot+'_C02.nc'),'r')
    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    # visx = conus05km.variables['x'][:] * satheight
    # visy = conus05km.variables['y'][:] * satheight
    # visxx,visyy = np.meshgrid(visx,visy)
    # vislons, vislats = p(visxx, visyy, inverse=True)
    # conus05km.close()

    crop = False
    if bounding_box[0] != -999:
        crop = True
        xmin, ymin = p(bounding_box[0], bounding_box[2])
        xmax, ymax = p(bounding_box[1], bounding_box[3])
        if sector == "RadC":

            xmin_ind, xmin = find_nearest(irx, xmin)
            xmax_ind, xmax = find_nearest(irx, xmax)
            ymin_ind, ymin = find_nearest(iry, ymin)
            ymax_ind, ymax = find_nearest(iry, ymax)

            startX = xmin_ind
            endX = xmax_ind
            startY = ymax_ind
            endY = ymin_ind
        else:  # assume it is a meso sector. This will also assume realtime=False,
            # as we don't want a static cropped domain for mesos in realtime.
            # get meso info from first dt in file
            dt = datetime.strptime(lines[0], "%Y%m%d-%H%M")
            try:
                visx, visy, irx, iry = get_meso_proj_info(dt, remote_dir, sector)
            except IndexError:
                logging.error(
                    "Couldn't locate meso data on "
                    + remote_dir
                    + "\
                      sector "
                    + sector
                    + " for time "
                    + dt.strftime("%Y%m%d-%H%M%S")
                )
                sys.exit(1)

            irxx, iryy = np.meshgrid(irx, iry)
            irxx = irxx[startY:endY, startX:endX]
            iryy = iryy[startY:endY, startX:endX]
            irlons, irlats = p(irxx, iryy, inverse=True)
            visxx = visxx[startY * 4 : endY * 4 + 1, startX * 4 : endX * 4 + 1]
            visyy = visyy[startY * 4 : endY * 4 + 1, startX * 4 : endX * 4 + 1]

    else:
        if sector == "RadC":
            # assume 'OCONUS' crop
            #    startX, endX = 1510, 2499
            #    startY, endY = 0, 970
            # assume 'CONUS' crop
            #    startX, endX = 55, 2300
            #    startY, endY = 32, 1040
            # southeast OCONUS crop
            startX, endX = 1510, 1910
            startY, endY = 450, 850
            crop = True

            xmin = irx[startX]
            xmax = irx[endX]
            ymin = iry[endY]
            ymax = iry[
                startY
            ]  # N.B.: startY is north, where y is greater. That's why it uses ymax
        else:  # assume we're doing the entire RadM1 or RadM2 sectors
            if realtime:
                pass  # projection info obtained from each dt in RT
            else:
                # get meso info from first dt in file
                dt = datetime.strptime(lines[0], "%Y%m%d-%H%M")
                try:
                    visx, visy, irx, iry = get_meso_proj_info(dt, remote_dir, sector)
                except IndexError:
                    logging.error(
                        "Couldn't locate meso data on "
                        + remote_dir
                        + "\
                        sector "
                        + sector
                        + " for time "
                        + dt.strftime("%Y%m%d-%H%M%S")
                    )
                    sys.exit(1)

            irxx, iryy = np.meshgrid(irx, iry)
            visxx, visyy = np.meshgrid(visx, visy)
            irlons, irlats = p(irxx, iryy, inverse=True)
            startY = endY = startX = endX = None

            # get x, y bounds for extent
            xmin = np.min(irx)
            xmax = np.max(irx)
            ymin = np.min(iry)
            ymax = np.max(iry)

    if crop:
        irlons = irlons[startY:endY, startX:endX]
        irlats = irlats[startY:endY, startX:endX]
        irxx = irxx[startY:endY, startX:endX]
        iryy = iryy[startY:endY, startX:endX]
        # vislons = vislons[startY*4:endY*4,startX*4:endX*4]
        # vislats = vislats[startY*4:endY*4,startX*4:endX*4]
        # visxx = visxx[startY*4:endY*4,startX*4:endX*4]
        # visyy = visyy[startY*4:endY*4,startX*4:endX*4]
    if sector == "RadM1" or sector == "RadM2":
        # glm crop info (always assumes goes_east conus for now)
        glmendY, glmstartX = earth_to_fgf(
            lat=[irlats[-1, 0]],
            lon=[irlons[-1, 0]],
            sub_lon_grid=-75.0,
            ccoff=-0.101332,
            lloff=0.128212,
        )
        glmstartX = int(np.round(glmstartX))
        glmendY = int(np.round(glmendY))
        glmstartY, glmendX = earth_to_fgf(
            lat=[irlats[0, -1]],
            lon=[irlons[0, -1]],
            sub_lon_grid=-75.0,
            ccoff=-0.101332,
            lloff=0.128212,
        )
        glmendX = int(np.round(glmendX))
        glmstartY = int(np.round(glmstartY))
    else:
        glmstartY = startY
        glmendY = endY
        glmstartX = startX
        glmendX = endX

    print("overhead:", time.time() - t00)

    line_cter = 0
    frame_cter = 0
    # if realtime, this is an infinite loop
    # otherwise, go through each line or time in filelist
    while line_cter < nlines or realtime:

        if realtime:
            while latest_filestr == filestr or filestr == "None":
                filestr, remoteDir, dt = find_latest(filelist, sector)
                if latest_filestr == filestr or filestr == "None":
                    logging.info("Already processed: " + latest_filestr)
                    logging.info("Sleeping 10 sec...")
                    time.sleep(10)
            # reset
            logging.info("Beginning to process: " + os.path.join(remoteDir, filestr))
            tstart = time.time()
            latest_filestr = filestr

        else:
            logging.info("processing " + lines[line_cter] + " UTC...")
            dt = datetime.strptime(lines[line_cter], "%Y%m%d-%H%M")

        input_list = [
            [] for _ in range(ninputs)
        ]  # list of lists of grids to be passed into generator. Currently, this won't work with the time dimension
        inp_cter = 0
        for (
            inp
        ) in (
            inputs
        ):  # this is for each input into the model (inputs is an OrderedDict...where order matters)
            for ch in inputs[inp]:  # this is for each channel in the input
                if ch.startswith("CH"):
                    # makes the 2D numpy array from data file
                    try:
                        meta = {}
                        if (
                            sector == "RadM1" or sector == "RadM2"
                        ) and realtime:  # also assumes we're not cropping
                            if ch == "CH13":
                                irx, iry, abi_data = get_abi_data(
                                    dt, ch[2:], remote_dir, metadata=meta
                                )
                                irxx, iryy = np.meshgrid(irx, iry)
                                irlons, irlats = p(irxx, iryy, inverse=True)
                        else:
                            abi_data = get_abi_data(
                                dt,
                                ch[2:],
                                remote_dir,
                                sector,
                                startY=startY,
                                endY=endY,
                                startX=startX,
                                endX=endX,
                                metadata=meta,
                            )
                    except IndexError as err:
                        logging.critical(traceback.format_exc())
                        logging.critical("Can't get input data for " + ch)
                        sys.exit(1)
                    if ch == "CH13":  # FIXME hardcoded
                        irdata = abi_data

                    # append the input data for use in the generator (generator does the scaling, as well)
                    input_list[inp_cter].append(abi_data)

            inp_cter += 1

        # end for one timestamp
        line_cter += 1

        logging.info("making predictions...")
        t0 = time.time()

        # find the NW corner points of patches to feed to generator
        corners, remainder_ny, remainder_nx = find_corners(
            mc["input_tuples"], input_list
        )

        try:
            gen = DataGeneratorPredict(
                input_list, mc, corners, batch_size=10, scalars=None
            )  # "batch size" determined internally
            # input_list gets modified with zero-padded grids
            # and is passed back
            new_preds = conv_model.predict(gen, workers=0, verbose=1)
        except ValueError as err:
            print(
                err
            )  # in realtime, this is an error possibly caused by certain input data not being available
            # will just move on to next scene
        else:

            logging.info("predict time: " + "{:.3f}".format(time.time() - t0))

            # get the number and shape of our predicted tiles
            ntiles, pred_ny, pred_nx = new_preds.shape  # new_preds = output tensor
            # instantiate output grid of predictions
            tmp_preds = np.zeros(irxx.shape)

            ir_corners = corners[-1]  # FIXME Assuming 2km corners will always be last
            y_corners, x_corners = zip(*ir_corners)
            y_corners = np.unique(y_corners)
            x_corners = np.unique(x_corners)

            # now fill the empty grid according to the corner pts and pred_ny, pred_nx
            # (this is basically reshaping our new_preds tensor)
            tile_cter = 0
            for yi, yc in enumerate(y_corners):
                for xi, xc in enumerate(x_corners):
                    if yc > 0 and yc != y_corners[-1] and xc == 0:  # left column
                        tmp_preds[
                            yc + 100 : yc + pred_ny, xc : xc + pred_nx
                        ] = new_preds[tile_cter][100:, :]
                    elif yc > 0 and xc == 0:  # "southwest" tile (use remainder_ny)
                        tmp_preds[-remainder_ny:, xc : xc + pred_nx] = new_preds[
                            tile_cter
                        ][-remainder_ny:, :]
                    elif yc == 0 and xc > 0 and xc != x_corners[-1]:  # top row
                        tmp_preds[
                            yc : yc + pred_ny, xc + 100 : xc + pred_nx
                        ] = new_preds[tile_cter][:, 100:]
                    elif yc == 0 and xc > 0:  # "northeast" tile (use remainder_nx)
                        tmp_preds[yc : yc + pred_ny, -remainder_nx:] = new_preds[
                            tile_cter
                        ][:, -remainder_nx:]
                    elif (
                        yc == y_corners[-1] and xc == x_corners[-1]
                    ):  # "southeast tile (use remainder_ny and remainder_nx)
                        tmp_preds[-remainder_ny:, -remainder_nx:] = new_preds[
                            tile_cter
                        ][-remainder_ny:, -remainder_nx:]
                    elif yc == 0 and xc == 0:  # "northwest" tile
                        tmp_preds[yc : yc + pred_ny, xc : xc + pred_nx] = new_preds[
                            tile_cter
                        ]
                    else:  # interior tile
                        tmp_preds[
                            yc + 100 : yc + pred_ny, xc + 100 : xc + pred_nx
                        ] = new_preds[tile_cter][100:, 100:]

                    tile_cter += 1

            # overwrite/rename our output preds grid
            new_preds = np.copy(tmp_preds)

            colors = ["#0000FF", "#00FFFF", "#00FF00", "#CC00FF"]
            rgb_colors = get_rgb_colors(colors)
            values = [0.1, 0.25, 0.5, 0.75]
            linewidths = [0.75, 1, 1, 1]
            stride = 4
            if realtime or make_json:
                print("making contours and json...")
                tj = time.time()
                fig = plt.figure(figsize=(15, 15))
                ax = plt.axes(projection=geoproj)

                # error fixing
                new_preds = np.ma.fix_invalid(new_preds, fill_value=0)
                new_preds[irlons < -130] = 0
                logging.info("max new_preds: " + "{:.3f}".format(np.max(new_preds)))

                contour_img = ax.contour(
                    irlons[0::stride, 0::stride],
                    irlats[0::stride, 0::stride],
                    new_preds[0::stride, 0::stride],
                    values,
                    transform=ccrs.PlateCarree(),
                )
                outjson = os.path.join(
                    outdir,
                    dt.strftime("%Y%m%d"),
                    "PLTG",
                    dt.strftime("PLTG_%Y%m%d_%H%M00.force.json"),
                )
                empty = False if (np.max(new_preds) > values[0]) else True
                write_json(
                    contour_img,
                    outjson,
                    values=values,
                    colors=rgb_colors,
                    re_upload=re_upload,
                    empty=empty,
                )
                logging.info("wrote out " + outjson)
                if realtime:
                    logging.info("total time: " + "{:.3f}".format(time.time() - tstart))

            elif savegrid:  # FIXME Not sure if this block works anymore...
                out_x = irx[hs["2km"] : -hs["2km"] : stride] / satheight
                out_y = iry[hs["2km"] : -hs["2km"] : stride] / satheight

                # error fixing
                new_preds = np.ma.fix_invalid(new_preds, fill_value=0)
                lons = np.array([vislons[c[0], c[1]] for c in viscenters]).reshape(
                    new_preds.shape
                )
                new_preds[lons < -130] = 0

                dims = OrderedDict()
                dims["x"] = len(out_x)
                dims["y"] = len(out_y)
                dims["g"] = 1

                dataset = {}
                dataset["x"] = {
                    "data": out_x,
                    "dims": ("x",),
                    "atts": {
                        "units": "rad",
                        "axis": "X",
                        "long_name": "GOES fixed grid projection x-coordinate",
                        "standard_name": "projection_x_coordinate",
                    },
                }
                dataset["y"] = {
                    "data": out_y,
                    "dims": ("y",),
                    "atts": {
                        "units": "rad",
                        "axis": "Y",
                        "long_name": "GOES fixed grid projection y-coordinate",
                        "standard_name": "projection_y_coordinate",
                    },
                }
                dataset["goes_imager_projection"] = {
                    "data": np.int32(1),
                    "dims": ("g",),
                    "atts": {},
                }
                conus2km = Dataset(
                    os.path.join(pltg, "src", "static", slot + ".nc"), "r"
                )
                for att in conus2km.variables["goes_imager_projection"].ncattrs():
                    dataset["goes_imager_projection"]["atts"][att] = conus2km.variables[
                        "goes_imager_projection"
                    ].getncattr(att)
                conus2km.close()
                dataset["intense_convection_probability"] = {
                    "data": new_preds,
                    "dims": ("y", "x"),
                    "atts": {"model": model},
                }
                atts = OrderedDict()
                atts["orbital_slot"] = slot
                atts["sector"] = sector
                atts["datetime"] = dt.strftime("%Y%m%d-%H%M%S UTC")

                netcdf = os.path.join(
                    outdir, dt.strftime("%Y%m%d/PLTG/PLTG_%Y%m%d-%H%M%S.netcdf")
                )
                status = utils.write_netcdf(
                    netcdf, dataset, dims, atts=atts, wait=True, gzip=False
                )

            else:  # make an image

                # get aspect ratio for IR data
                default_imgsize = 10  # inches
                try:
                    x_range = endX - startX
                    y_range = endY - startY
                except TypeError:  # i.e., endX, etc. are 'None'
                    y_range, x_range = irxx.shape

                asp = y_range / x_range
                if asp > 1:
                    fig_width_in = int(default_imgsize * asp ** (-1))
                    fig_height_in = int(default_imgsize)
                else:
                    fig_width_in = int(default_imgsize)
                    fig_height_in = int(default_imgsize * asp)

                if glmdir:
                    # find best file
                    glmfile = glob.glob(
                        dt.strftime(
                            os.path.join(
                                glmdir,
                                "%Y%m%d",
                                "GLM",
                                "goes_east",
                                "agg",
                                "%Y%m%d-%H%M*.netcdf*",
                            )
                        )
                    )
                    if not (glmfile):
                        for mm in [1, 2, 3, 4, 5]:
                            newdt = dt - timedelta(minutes=mm)
                            glmfile = glob.glob(
                                newdt.strftime(
                                    os.path.join(
                                        glmdir,
                                        "%Y%m%d",
                                        "GLM",
                                        "goes_east",
                                        "agg",
                                        "%Y%m%d-%H%M*.netcdf*",
                                    )
                                )
                            )
                            if glmfile:
                                break
                    if glmfile:
                        glmdata = {"flash_extent_density": -1}
                        ss = read_netcdf(glmfile[0], data=glmdata)
                        imgdata = glmdata["flash_extent_density"][
                            glmstartY:glmendY, glmstartX:glmendX
                        ]

                        vmin = 0
                        vmax = 80
                        inc = 10
                        cmap = plt.get_cmap("autumn_r")
                        cbar_label = "GLM flash extent density [fl/(5-min)]"
                        out_prefix = "GLMFED"
                else:
                    # assume an IR image
                    imgdata = frames["2km"][-1]
                    vmin = 180
                    vmax = 310
                    inc = 10
                    cmap = irw_ctc()
                    cbar_label = "Channel " + channel + " brightness temperature [K]"
                    out_prefix = "C" + channel

                fig = plt.figure(figsize=((fig_width_in, fig_height_in)))
                ax = fig.add_axes([0, 0, 1, 1], projection=geoproj)
                # set extent of image
                ax.set_extent([xmin, xmax, ymin, ymax], crs=geoproj)
                ax.add_feature(cfeature.BORDERS)
                ax.add_feature(state_borders, edgecolor="tan")

                #  pcimg = ax.pcolormesh(visxx, visyy, np.sqrt(visdata), vmin=0, vmax=1, transform=geoproj, cmap=plt.get_cmap('Greys_r'))
                #  imgd = ma.masked_less(zoom(imgdata,4,order=1),1); irxz = zoom(irxx,4,order=1); iryz = zoom(iryy,4,order=1)
                #  pcimg = ax.pcolormesh(irxz, iryz, imgd, vmin=vmin, vmax=vmax, transform=geoproj, cmap=cmap, alpha=0.65, linewidth=0, antialiased=True)

                fs = 12
                irimg = ax.pcolormesh(
                    irxx, iryy, irdata, vmin=180, vmax=310, cmap=plt.get_cmap("Greys")
                )
                cbaxes = fig.add_axes(
                    [0.01, -0.05, 0.3, 0.03]
                )  # left bottom width height
                cbar = plt.colorbar(
                    irimg,
                    orientation="horizontal",
                    pad=0.01,
                    extendfrac=0,
                    drawedges=0,
                    ax=ax,
                    ticks=np.arange(180, 301, 40),
                    cax=cbaxes,
                )
                cbar.set_label("ABI 10.3-$\mu$m BT [K]", fontsize=fs)
                cbar.ax.tick_params(labelsize=fs)

                glmimg = ax.pcolormesh(
                    irxx,
                    iryy,
                    ma.masked_less(imgdata, 0.1),
                    vmin=vmin,
                    vmax=vmax,
                    cmap=plt.get_cmap("autumn_r"),
                    linewidth=0,
                    antialiased=True,
                    alpha=0.5,
                )
                cbaxes2 = fig.add_axes(
                    [0.4, -0.05, 0.3, 0.03]
                )  # left bottom width height
                cbar2 = plt.colorbar(
                    glmimg,
                    orientation="horizontal",
                    pad=0.01,
                    extendfrac=0,
                    drawedges=0,
                    ax=ax,
                    ticks=np.arange(10, 71, 20),
                    cax=cbaxes2,
                )
                cbar2.set_label("Max GLM FED [fl (5 min)$^{-1}$]", fontsize=fs)
                cbar2.ax.tick_params(labelsize=fs)
                cbar2.set_alpha(1.0)
                cbar2.draw_all()

                # contours
                labels = [str(int(v * 100)) + "%" for v in values]

                contour_img = ax.contour(
                    irxx[0::stride, 0::stride],
                    iryy[0::stride, 0::stride],
                    new_preds[0::stride, 0::stride],
                    values,
                    colors=colors,
                    linewidths=linewidths,
                )  # , cmap=plt.get_cmap('cividis'))
                custom_lines = [
                    Line2D([0], [0], color=c, lw=lw * 2)
                    for c, lw in zip(colors, linewidths)
                ]
                if len(labels) > 3:
                    ncol = 2
                    vert_offset_lab = -0.17
                    vert_offset_leg = -0.14
                    horiz_offset_leg = 0.95
                else:
                    ncol = len(labels)
                    vert_offset_lab = -0.11
                    vert_offset_leg = -0.08
                    horiz_offset_leg = 1.0
                leg2 = ax.legend(
                    custom_lines,
                    labels,
                    fontsize=fs,
                    loc="lower right",
                    handlelength=1.5,
                    ncol=ncol,
                    bbox_to_anchor=(horiz_offset_leg, vert_offset_leg),
                )
                ctext = plt.text(
                    0.66,
                    vert_offset_lab,
                    "Probability of lightning in 60 min",
                    fontsize=fs,
                    transform=ax.transAxes,
                )

                goes_position = meta["goes_position"]
                scene_id = meta["scene_id"]
                timestamp = dt.strftime("%Y%m%d-%H%M")
                ax.set_title(
                    (" ").join([goes_position, scene_id, timestamp + " UTC"]),
                    fontsize=fs + 2,
                )

                fig.savefig(
                    os.path.join(
                        outdir, ("_").join([out_prefix, scene_id, timestamp + ".png"])
                    ),
                    bbox_inches="tight",
                )
                plt.close()

        # back up frame and frame_cter
        # frame_cter -= 1
        # if(nframes > 1):
        #  for k in frames:
        #    if(isinstance(frames[k],np.ndarray)):
        #      frames[k][0:nframes-1] = frames[k][1:nframes]


################################################################################################################################################################################
if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Takes a list of file sequence or rabbitMQ real-time list of GOES data and \
                                                will make predictions of lightning in the next 30 min. Can make images, jsons, or grids."
    )
    parser.add_argument(
        "filelist",
        help="A list of channel filenames. Channels should be in order that model expects them. See chan_filelist.txt for example.\
                                 If realtime is set, then this argument is the listened log output from amqpfind.",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "model", help="CNN model used for predictions.", type=str, nargs=1
    )
    parser.add_argument(
        "model_config_file",
        help="This file contains mean and SD of channels (a pickle file).",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-rd",
        "--remote_dir",
        help='What partition to use. Default = "/arcdata/goes/grb/goes16/"',
        type=str,
        nargs=1,
        default=["/arcdata/goes/grb/goes16/"],
    )
    parser.add_argument(
        "-sec",
        "--sector",
        help="RadC, RadM1, or RadM2. Default = 'RadC'",
        default=["RadC"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-pv",
        "--prob_vals",
        help="probability values to contour (0 to 1). Default = 0.25 0.5 0.9",
        type=float,
        nargs=3,
        default=[0.25, 0.5, 0.9],
    )
    parser.add_argument(
        "-g",
        "--glmdir",
        help="root glm dir. Default: /apollo/grain/saved_probsevere_data/lightning/ ...will append yyyymmdd/GLM/goes_east/agg/",
        type=str,
        nargs=1,
        default=["/apollo/grain/saved_probsevere_data/lightning/"],
    )
    parser.add_argument(
        "-b",
        "--bounding_box",
        help="Bounding box of image. SWlon, NElon, SWlat, NElat. Default is entire image.",
        default=[-999],
        type=float,
        nargs=4,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="Root output directory. Default=$PWD/OUTPUT/",
        default=[os.path.join(os.environ["PWD"], "OUTPUT/")],
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-a",
        "--alpha",
        help="alpha for IR image in sandwich. Default is 0.25",
        default=[0.25],
        type=float,
        nargs=1,
    )
    parser.add_argument(
        "-sl",
        "--slot",
        help="orbital slot...GOES_East or GOES_West. Only used if realtime=True",
        default=["GOES_East"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-ru",
        "--re_upload",
        help="Full path to re_upload executable. Will execute after json is written.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-sg",
        "--savegrid",
        help="Save a netcdf of the ICP to outdir. Only in offline mode right now.",
        action="store_true",
    )
    parser.add_argument(
        "-mj",
        "--make_json",
        help="Save a json to outdir. --realtime does this by default.",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--realtime",
        help="Set for realtime processing. filelist arg is the watched file from ampqfind.",
        action="store_true",
    )

    args = parser.parse_args()

    logDict = logging_def(logfile=None)
    # do this regardless of realtime/offline, so we get formatted stdout messages
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)

    predict_ltg(
        args.filelist[0],
        args.model[0],
        args.model_config_file[0],
        outdir=args.outdir[0],
        alpha=args.alpha[0],
        bounding_box=args.bounding_box,
        remote_dir=args.remote_dir[0],
        values=args.prob_vals,
        glmdir=args.glmdir[0],
        sector=args.sector[0],
        slot=args.slot[0],
        realtime=args.realtime,
        re_upload=args.re_upload[0],
        savegrid=args.savegrid,
        make_json=args.make_json,
    )
