import matplotlib

matplotlib.use("Agg")
import glob
import shutil
import os, sys
import numpy as np
import zarr
import netCDF4
import torch, os
import torch.nn as nn

# from torch.utils.tensorboard import SummaryWriter
import time
import pickle
import nvidia_smi
from lightningcast.pytorchtools import EarlyStopping
from lightningcast import utils
import collections
from lightningcast import torch_losses
from lightningcast import eval_torch_cnn
from lightningcast import torch_generator
import argparse
from lightningcast import sat_zen_angle

# os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID";
# os.environ["CUDA_VISIBLE_DEVICES"]="2"


class Print(nn.Module):
    def forward(self, x):
        print(x.size())
        return x


# -------------------------------------------------------------------------------------------------#
def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group["lr"]


# -------------------------------------------------------------------------------------------------#
def count_parameters(model):
    print(
        "num parameters = ",
        sum(p.numel() for p in model.parameters() if p.requires_grad),
    )


# -------------------------------------------------------------------------------------------------#
class accuracy_metric:
    def __init__(self):
        self.total = 0
        self.correct = 0
        self.data = []

    def add(self, outputs, labels):
        _, predicted = torch.max(outputs.data, 1)
        self.total += labels.size(1) * labels.size(2)  # 1 and 2 b/c of UNET
        self.correct += (predicted == labels).sum().item()

    def value(self):
        val = self.correct / self.total if self.total else 0
        self.data.extend([val])
        return val


# -------------------------------------------------------------------------------------------------#
class running_mean:
    def __init__(self):
        self.total = 0
        self.count = 0

    def add(self, value):
        self.total += value
        self.count += 1

    def value(self):
        return self.total / self.count if self.count else 0


# -------------------------------------------------------------------------------------------------------#

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config_file",
        help="Enter a config pkl file. Default is to excecute torch_config.py",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-m",
        "--model_file",
        help="Model file. Resume learning with this file.",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-p",
        "--partition_file",
        help="Only valid if no --data (using a generator). Pickle file containing dict of train/val \
                                              partition netcdfs  and labels. Will also look for one in --outidr",
        nargs=1,
        type=str,
    )
    args = parser.parse_args()

    if args.config_file:
        config_file = args.config_file[0]
        print("opening " + config_file)
        config = pickle.load(open(config_file, "rb"))
    else:
        from lightningcast import torch_config

        config = torch_config.torch_config()
    ngpu = config["ngpu"]
    nclass = config["nclass"]

    # CUDA for PyTorch
    if ngpu > 0 and torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        ngpu = 0
        device = torch.device("cpu")

    # get LR
    if isinstance(config["learning_rate"], list):
        LR = config["learning_rate"][-1]
    else:
        LR = config["learning_rate"]

    # outdir
    outdir = config["outdir"]
    utils.mkdir_p(outdir)

    # Restore model, if necessary. This means we're starting from a checkpoint,
    # so we also want to restore the latest epoch, loss value, and learning rate.
    if args.model_file:
        model_file = args.model_file[0]
        print("opening " + model_file)
        sys.path.insert(
            0, os.path.dirname(model_file)
        )  # Use torch_models.py in model directory, not the one in src,
        # as model def might have changed. This command just puts the
        # model directory at front of the path
        assert os.path.isdir(os.path.dirname(model_file))
        from lightningcast import torch_models

        torch_model = getattr(torch_models, config["cnn"])
        model = torch_model(config)
        mdict = torch.load(model_file, map_location=device)
        model.load_state_dict(mdict["model_state_dict"])
        try:
            start_epoch = config["latest_epoch"]
            best_loss = config["best_loss"]
        except KeyError:  # if we get here, it means we started with an existing model, but
            # we're using a new config to continue training.
            start_epoch = 0
            best_loss = None

    else:
        # start with a new model
        from lightningcast import torch_models

        torch_model = getattr(torch_models, config["cnn"])
        model = torch_model(config)
        # copy the model program/script to the outdir, as we don't want to lose the model definition
        shutil.copy("torch_models.py", outdir)
        # defaults when starting fresh
        start_epoch = 0
        best_loss = None

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # device=torch.device('cpu')
    print(f"Using device = {device}")
    if device.type == "cuda":
        nvidia_smi.nvmlInit()
        handle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
        print("GPU type =", torch.cuda.get_device_name(0))
    # what is this line for?
    torch.backends.cudnn.benchmark = True

    datadir = config["data"]
    scalar_vars = config["scalar_vars"]

    # static reduced-res (RR) lats, lons, satzen
    if (
        "lat" in scalar_vars
        or "lon" in scalar_vars
        or "satzen" in scalar_vars
        or "solzen" in scalar_vars
    ):
        _, _, _, lon, lat = utils.get_area_definition(
            "static/GOES_East.nc", return_latlons=True
        )
        srrf = config["scalar_reduce_res_fac"]
        lonRR = lon[srrf // 2 :: srrf, srrf // 2 :: srrf]
        latRR = lat[srrf // 2 :: srrf, srrf // 2 :: srrf]
        satzenRR = (
            sat_zen_angle.sat_zen_angle(xlat=latRR, xlon=lonRR, satlat=0, satlon=-75.0)
            if ("satzen" in scalar_vars)
            else None
        )
    else:
        latRR = lonRR = satzenRR = None

    # Formulating generators
    DG = getattr(torch_generator, config["datagen"])

    if config["datagen"] == "DatasetZARR_UNET":
        print("Making training and validaiton zarr ind lists...")
        zf = zarr.open(config["data"])
        # get the inds
        if args.partition_file:
            part_file = pickle.load(open(args.partition_file[0], "rb"))
            partition = part_file["partition"]
            labels = part_file["labels"]
            val_ind = partition["validation"]
            train_ind = partition["train"]
        elif os.path.isfile(os.path.join(outdir, "part_labs.pkl")):
            part_file = pickle.load(open(os.path.join(outdir, "part_labs.pkl"), "rb"))
            partition = part_file["partition"]
            labels = part_file["labels"]
            val_ind = partition["validation"]
            train_ind = partition["train"]
        else:
            partition = {}
            labels = {}

            days = zf.ABI_date[:]
            ind = np.arange(len(days))
            val_mask = np.isin(days, config["val_dates"])
            if len(config["train_dates"]):
                train_mask = np.isin(days, config["train_dates"])
            else:
                train_mask = np.invert(val_mask)
            val_ind = ind[val_mask]
            train_ind = ind[train_mask]

            partition["train"] = train_ind
            partition["validation"] = val_ind

            print("Saving " + os.path.join(outdir, "part_labs.pkl..."))
            pickle.dump(
                {
                    "partition": partition,
                    "train_mask": train_mask,
                    "val_mask": val_mask,
                    "labels": labels,
                },
                open(os.path.join(outdir, "part_labs.pkl"), "wb"),
                protocol=4,
            )
        n_tsamples = len(train_ind)
        n_vsamples = len(val_ind)
        training_set = DG(train_ind, config, zarr=zf)
        validation_set = DG(val_ind, config, zarr=zf)

    else:
        # Non-zarr. Using netcdfs.
        # if we're using a list of dates
        if len(config["train_dates"]):
            val_files = []
            for vd in config["val_dates"]:
                val_files += glob.glob(os.path.join(datadir, str(vd), "*.netcdf"))
            val_files = np.sort(val_files)
            n_vsamples = len(val_files)
            validation_set = DG(val_files, config)

            train_files = []
            for td in config["train_dates"]:
                train_files += glob.glob(os.path.join(datadir, str(td), "*.netcdf"))
            train_files = np.sort(train_files)
            n_tsamples = len(train_files)
            training_set = DG(train_files, config)
        else:
            # We're using samplelists (pickles containing lists of dicts)
            train_list = pickle.load(open(config["train_list"], "rb"))
            training_set = DG(
                train_list, config, lats=latRR, lons=lonRR, satzen=satzenRR
            )
            val_list = pickle.load(open(config["val_list"], "rb"))
            validation_set = DG(
                val_list, config, lats=latRR, lons=lonRR, satzen=satzenRR
            )
            n_tsamples = len(train_list)
            n_vsamples = len(val_list)

    print("Number of validation samples:", n_vsamples)
    print("Number of training samples:", n_tsamples)
    print("Ratio of val/all samples:", n_vsamples / (n_vsamples + n_tsamples))

    bs = config["batchsize"]
    nworkers = config["nworkers"]

    # ngpu = torch.cuda.device_count()
    if ngpu > 1:
        print("Let's use", ngpu, "GPUs!")
        model = nn.DataParallel(model)
    model.to(device)
    print(model)
    count_parameters(model)

    assert bs >= ngpu

    # Parameters
    params = {
        "batch_size": bs,  # best to use a batch_size that is a multiple of your ngpus
        "shuffle": True,
        "num_workers": nworkers,
    }

    training_generator = torch.utils.data.DataLoader(training_set, **params)
    validation_generator = torch.utils.data.DataLoader(validation_set, **params)

    # Configure
    if config["optimizer"].upper() == "RADAM":
        from lightningcast.optimizers import RAdam

        optimizer = RAdam(model.parameters(), lr=LR)
    else:
        optimizer = torch.optim.AdamW(model.parameters(), lr=LR)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer,
        "min",
        patience=config["rlr_patience"],
        factor=config["rlr_factor"],
        threshold=0.0001,
        cooldown=0,
    )
    early_stopping = EarlyStopping(
        patience=config["es_patience"],
        verbose=True,
        path=outdir,
        ngpu=ngpu,
        best_score=best_loss,
    )

    loss_str = config["loss_fcn"]
    if loss_str == "BCE":
        loss_fcn = nn.BCELoss()
    elif (
        loss_str == "BCEWithLogitsLoss"
    ):  # pos_weight should be one number, for the positive class. Could also compute on a per-batch basis
        # and use functional api (torch.nn.functional.binary_cross_entropy_with_logits) with pos_weight.
        # See https://discuss.pytorch.org/t/should-i-use-nn-bcewithlogitsloss-or-cross-entropy-loss-for-segmentation/79476/4
        # BCEWithLogitsLoss combines Sigmoid with loss, which is numerically more stable.
        loss_fcn = nn.BCEWithLogitsLoss(
            pos_weight=torch.tensor([100], device=device, dtype=torch.float)
        )
        label_dtype = torch.float
    elif loss_str == "CrossEntropyLoss":
        loss_fcn = nn.CrossEntropyLoss(
            weight=torch.tensor([0.5, 100], device=device, dtype=torch.float)
        )
        label_dtype = torch.long
    else:
        loss_fcn = getattr(torch_losses, config["loss_fcn"])()
        label_dtype = torch.float

    # writer    = SummaryWriter()
    acc_list = []

    pickle.dump(config, open(os.path.join(outdir, "model_config.pkl"), "wb"))

    max_epochs = config["nepoch"]
    # ---Training and validation loop---#
    for epoch in range(start_epoch + 1, max_epochs):
        # Training
        model.train()
        train_acc = accuracy_metric()
        train_loss = running_mean()
        gpu = running_mean()
        t0 = time.time()
        end = time.time()
        curr_lr = optimizer.param_groups[0]["lr"]
        for local_batch, local_labels in training_generator:
            # Transfer to GPU
            local_batch_list = []
            for lb in local_batch:  # add all X inputs to GPU
                local_batch_list.append(lb.to(device, dtype=torch.float))
            # add labels to GPU
            local_labels = local_labels.to(device, dtype=label_dtype)

            # Model computations
            # Zero the gradients
            optimizer.zero_grad()
            # forward, backward, optimize

            outputs = model(local_batch_list)
            loss = loss_fcn(outputs, local_labels)
            loss.backward()
            optimizer.step()

            train_loss.add(loss.item())
            train_acc.add(outputs, local_labels)
            if device.type == "cuda":
                gpu.add(nvidia_smi.nvmlDeviceGetUtilizationRates(handle).gpu)
            end = time.time()

        # Validation
        with torch.set_grad_enabled(False):
            model.eval()
            val_loss = running_mean()
            val_acc = accuracy_metric()
            for local_batch, local_labels in validation_generator:
                # Transfer to GPU
                local_batch_list = []
                for lb in local_batch:  # add all X inputs to GPU
                    local_batch_list.append(lb.to(device, dtype=torch.float))
                # add labels to GPU
                local_labels = local_labels.to(device, dtype=label_dtype)

                # local_batch, local_labels = local_batch.to(device,dtype=torch.float), local_labels.to(device,dtype=torch.float)

                # Model computations
                outputs = model(local_batch_list)
                loss = loss_fcn(outputs, local_labels)
                val_loss.add(loss.item())
                val_acc.add(outputs, local_labels)

        # DISPLAY PROGRESS
        print(
            f"epoch={epoch:3d} dt={time.time()-t0:.3f}s gpu ={gpu.value():.0f}% training loss={train_loss.value():.5f} val loss ={val_loss.value():.5f} LR ={curr_lr:.7f} train_acc={train_acc.value():.3f} val_acc={val_acc.value():.3f} "
        )
        # writer.add_scalar('Loss/val'      , val_loss.value()  , epoch)
        # writer.add_scalar('Loss/train'    , train_loss.value(), epoch)
        # writer.add_scalar('Accuracy/val'  , val_acc.value()   , epoch)
        # writer.add_scalar('Accuracy/train', train_acc.value() , epoch)

        acc_list.extend([val_acc.value()])
        # check learning rate; reduce if necessary
        scheduler.step(val_loss.value())

        # early stopping and model checkpoint
        early_stopping(val_loss.value(), model, optimizer, epoch)
        if early_stopping.early_stop:
            print("Early stopping")
            break

        # add to config, in case we need to restart from checkpoint
        config["latest_epoch"] = epoch
        config["best_loss"] = early_stopping.best_score
        if isinstance(config["learning_rate"], list):
            config["learning_rate"].append(optimizer.param_groups[0]["lr"])
        else:  # it's a float...make a list
            config["learning_rate"] = [
                config["learning_rate"],
                optimizer.param_groups[0]["lr"],
            ]
        # save
        pickle.dump(config, open(os.path.join(outdir, "model_config.pkl"), "wb"))

    # --Evaluation--#

    # preallocate
    ny = config["input_tuples"][-1][-2]
    nx = config["input_tuples"][-1][-1]
    preds = np.empty((n_vsamples, 1, ny, nx), dtype=np.float16)
    labels = np.empty((n_vsamples, 1, ny, nx), dtype=np.float16)

    remainder = n_vsamples % bs

    params = {"batch_size": bs, "shuffle": False, "num_workers": 8}

    validation_generator = torch.utils.data.DataLoader(validation_set, **params)

    starttime = time.time()
    with torch.set_grad_enabled(False):
        cter = 0
        for local_batch, local_labels in validation_generator:  # for each sample
            # Transfer to device (GPU or CPU)
            local_batch_list = []
            for lb in local_batch:  # for each input of each sample
                local_batch_list.append(lb.to(device, dtype=torch.float))

            # make local_labels numpy and add to array
            try:
                labels[cter : cter + bs] = local_labels.cpu().numpy()
            except ValueError:
                labels[cter : cter + remainder] = local_labels.cpu().numpy()

            # Model computations
            outputs = model(local_batch_list).cpu().numpy()
            if nclass == 2:  # FIXME this block might needs some work...
                print(outputs.shape)
                try:
                    preds[cter : cter + bs] = outputs[
                        :, 1, ...
                    ]  # used with cross entropy...should be getting only pos class
                except ValueError:
                    preds[cter : cter + remainder] = outputs[:, 1, ...]
            else:
                try:
                    preds[cter : cter + bs] = outputs
                except ValueError:
                    preds[cter : cter + remainder] = outputs
            cter += 1

    print(
        "Made predictions in " + "{:.0f}".format(time.time() - starttime) + " seconds"
    )
    torch.cuda.empty_cache()

    stride = 4

    preds = np.squeeze(preds, axis=1)
    preds = preds[:, ::stride, ::stride]  # stride in both x and y
    preds = preds.flatten().astype(
        np.float32
    )  # float32 precision needed for validation routines (e.g., MSE)

    labels = np.squeeze(labels, axis=1)
    labels = labels[:, ::stride, ::stride]  # stride in both x and y
    labels = labels.flatten().astype(
        np.float32
    )  # float32 precision needed for validation routines (e.g., MSE)

    if (
        preds.nbytes / (1024**3) > 15
    ):  # If data > 10 GB, just score probs that are > than some threshold.
        # We do this because in the verification function, we seem to run out
        # of RAM, even with 180GB, if the data are 15GB or more.
        good_ind = np.where(preds > 0.001)
        print(len(preds))
        preds = preds[good_ind]
        print(len(preds))
        labels = labels[good_ind]

    # binarize
    print("binarizing...")
    labels[labels >= config["binarize"]] = 1
    labels[labels < config["binarize"]] = 0

    scores_dict = eval_torch_cnn.verification(preds, labels, outdir)

    pickle.dump(scores_dict, open(os.path.join(outdir, "scores_dict.pkl"), "wb"))
    of = open(os.path.join(outdir, "verification_scores.txt"), "a")
    try:
        of.write("data source: " + config["data"] + "\n")
    except TypeError:
        pass
    of.close()
